from __future__ import annotations
from typing import Any
from ..fable_library.option import (some, map)
from ..fable_library.types import Array
from .fable_py import (get_static_property_helpers, try_get_static_property_helper)
from .property_helper import PropertyHelper

def get_static_properties(o: Any=None) -> Array[PropertyHelper]:
    return get_static_property_helpers(o)


def try_get_static_property_info(o: Any, prop_name: str) -> PropertyHelper | None:
    return try_get_static_property_helper(o, prop_name)


def try_set_property_value(o: Any, prop_name: str, value: Any=None) -> bool:
    match_value: PropertyHelper | None = try_get_static_property_info(o, prop_name)
    (pattern_matching_result, property_1) = (None, None)
    if match_value is not None:
        if match_value.IsMutable:
            pattern_matching_result = 0
            property_1 = match_value

        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        property_1.SetValue(o, value)
        return True

    elif pattern_matching_result == 1:
        return False



def try_get_property_value(o: Any, prop_name: str) -> Any | None:
    try: 
        match_value: PropertyHelper | None = try_get_static_property_info(o, prop_name)
        return None if (match_value is None) else some(match_value.GetValue(o))

    except Exception as match_value_1:
        return None



def try_get_property_value_as(o: Any, prop_name: str) -> Any | None:
    try: 
        def mapping(v: Any=None) -> Any:
            return v

        return map(mapping, try_get_property_value(o, prop_name))

    except Exception as match_value:
        return None



def remove_property(o: Any, prop_name: str) -> bool:
    match_value: PropertyHelper | None = try_get_static_property_info(o, prop_name)
    (pattern_matching_result, property_1) = (None, None)
    if match_value is not None:
        if match_value.IsMutable:
            pattern_matching_result = 0
            property_1 = match_value

        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        property_1.RemoveValue(o)
        return True

    elif pattern_matching_result == 1:
        return False



__all__ = ["get_static_properties", "try_get_static_property_info", "try_set_property_value", "try_get_property_value", "try_get_property_value_as", "remove_property"]




from __future__ import annotations
from dataclasses import dataclass
from typing import Any
from ..fable_library.option import default_arg
from ..fable_library.reflection import (TypeInfo, union_type, string_type, array_type, record_type, obj_type, option_type)
from ..fable_library.types import (Array, Union, Record)

def _expr3686() -> TypeInfo:
    return union_type("ARCtrl.Contract.DTOType", [], DTOType, lambda: [[], [], [], [], [], [], [], [], [], [], [], []])


class DTOType(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["ISA_Assay", "ISA_Study", "ISA_Workflow", "ISA_Run", "ISA_Investigation", "ISA_Datamap", "JSON", "Markdown", "CWL", "YAML", "PlainText", "Cli"]


DTOType_reflection = _expr3686

def _expr3687() -> TypeInfo:
    return record_type("ARCtrl.Contract.CLITool", [], CLITool, lambda: [("Name", string_type), ("Arguments", array_type(string_type))])


@dataclass(eq = False, repr = False, slots = True)
class CLITool(Record):
    Name: str
    Arguments: Array[str]
    @staticmethod
    def create(name: str, arguments: Array[str]) -> CLITool:
        return CLITool(name, arguments)


CLITool_reflection = _expr3687

def _expr3688() -> TypeInfo:
    return union_type("ARCtrl.Contract.DTO", [], DTO, lambda: [[("Item", obj_type)], [("Item", string_type)], [("Item", CLITool_reflection())]])


class DTO(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Spreadsheet", "Text", "CLITool"]


DTO_reflection = _expr3688

def DTO__get_isSpreadsheet(this: DTO) -> bool:
    if this.tag == 0:
        return True

    else: 
        return False



def DTO__get_isText(this: DTO) -> bool:
    if this.tag == 1:
        return True

    else: 
        return False



def DTO__get_isCLITool(this: DTO) -> bool:
    if this.tag == 2:
        return True

    else: 
        return False



def DTO__AsSpreadsheet(this: DTO) -> Any:
    if this.tag == 0:
        return this.fields[0]

    else: 
        raise Exception("Not a spreadsheet")



def DTO__AsText(this: DTO) -> str:
    if this.tag == 1:
        return this.fields[0]

    else: 
        raise Exception("Not text")



def DTO__AsCLITool(this: DTO) -> CLITool:
    if this.tag == 2:
        return this.fields[0]

    else: 
        raise Exception("Not a CLI tool")



def _expr3689() -> TypeInfo:
    return record_type("ARCtrl.Contract.Contract", [], Contract, lambda: [("Operation", string_type), ("Path", string_type), ("DTOType", option_type(DTOType_reflection())), ("DTO", option_type(DTO_reflection()))])


@dataclass(eq = False, repr = False, slots = True)
class Contract(Record):
    Operation: str
    Path: str
    DTOType: DTOType | None
    DTO: DTO | None
    @staticmethod
    def create(op: str, path: str, dto_type: DTOType | None=None, dto: DTO | None=None) -> Contract:
        return Contract(op, path, dto_type, dto)

    @staticmethod
    def create_create(path: str, dto_type: DTOType, dto: DTO | None=None) -> Contract:
        return Contract("CREATE", path, dto_type, dto)

    @staticmethod
    def create_update(path: str, dto_type: DTOType, dto: DTO) -> Contract:
        return Contract("UPDATE", path, dto_type, dto)

    @staticmethod
    def create_delete(path: str) -> Contract:
        return Contract("DELETE", path, None, None)

    @staticmethod
    def create_read(path: str, dto_type: DTOType) -> Contract:
        return Contract("READ", path, dto_type, None)

    @staticmethod
    def create_execute(dto: CLITool, path: str | None=None) -> Contract:
        return Contract("EXECUTE", default_arg(path, ""), DTOType(11), DTO(2, dto))

    @staticmethod
    def create_rename(old_path: str, new_path: str) -> Contract:
        return Contract("RENAME", old_path, None, DTO(1, new_path))


Contract_reflection = _expr3689

__all__ = ["DTOType_reflection", "CLITool_reflection", "DTO_reflection", "DTO__get_isSpreadsheet", "DTO__get_isText", "DTO__get_isCLITool", "DTO__AsSpreadsheet", "DTO__AsText", "DTO__AsCLITool", "Contract_reflection"]




from __future__ import annotations
from dataclasses import dataclass
from typing import (Any, TypeVar)
from ....fable_library.reflection import (TypeInfo, string_type, record_type)
from ....fable_library.seq import map
from ....fable_library.types import Record
from ....fable_library.util import (to_enumerable, IEnumerable_1)
from ....thoth_json_core.types import (IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

def _expr1815() -> TypeInfo:
    return record_type("ARCtrl.Json.ROCrateContext.Sample.IContext", [], IContext, lambda: [("sdo", string_type), ("arc", string_type), ("Sample", string_type), ("ArcSample", string_type), ("name", string_type), ("characteristics", string_type), ("factor_values", string_type), ("derives_from", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class IContext(Record):
    sdo: str
    arc: str
    Sample: str
    ArcSample: str
    name: str
    characteristics: str
    factor_values: str
    derives_from: str

IContext_reflection = _expr1815

def _arrow1822(__unit: None=None) -> IEncodable:
    class ObjectExpr1816(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
            return helpers.encode_string("http://schema.org/")

    class ObjectExpr1817(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
            return helpers_1.encode_string("https://bioschemas.org/")

    class ObjectExpr1818(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
            return helpers_2.encode_string("bio:Sample")

    class ObjectExpr1819(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
            return helpers_3.encode_string("sdo:name")

    class ObjectExpr1820(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
            return helpers_4.encode_string("bio:additionalProperty")

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("sdo", ObjectExpr1816()), ("bio", ObjectExpr1817()), ("Sample", ObjectExpr1818()), ("name", ObjectExpr1819()), ("additionalProperties", ObjectExpr1820())])
    class ObjectExpr1821(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_5))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_5.encode_object(arg)

    return ObjectExpr1821()


context_jsonvalue: IEncodable = _arrow1822()

__all__ = ["IContext_reflection", "context_jsonvalue"]




from __future__ import annotations
from typing import Any
from ...arctrl_core.comment import Comment
from ...arctrl_core.Helper.collections_ import (Option_fromSeq, ResizeArray_map)
from ...arctrl_core.Helper.orcid import try_get_orcid_number
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.person import Person
from ...arctrl_json.ROCrate.ldnode import (decoder as decoder_1, encoder)
from ...arctrl_rocrate.ldcontext import LDContext
from ...arctrl_rocrate.ldobject import (LDNode, LDGraph)
from ...arctrl_rocrate.LDTypes.organization import LDOrganization
from ...arctrl_rocrate.LDTypes.person import LDPerson
from ...fable_library.option import map
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.result import FSharpResult_2
from ...fable_library.seq import (is_empty, filter)
from ...fable_library.string_ import (to_text, printf)
from ...fable_library.types import (Array, to_string as to_string_1)
from ...thoth_json_python.decode import Decode_fromString
from ...thoth_json_python.encode import to_string
from .basic import (BaseTypes_composeDefinedTerm_ZDED3A0F, BaseTypes_decomposeDefinedTerm_Z2F770004)

def _expr3919() -> TypeInfo:
    return class_type("ARCtrl.Conversion.PersonConversion", None, PersonConversion)


class PersonConversion:
    ...

PersonConversion_reflection = _expr3919

def PersonConversion_get_orcidKey(__unit: None=None) -> str:
    return "ORCID"


def PersonConversion_composeAffiliation_Z721C83C5(affiliation: str) -> LDNode:
    try: 
        match_value: FSharpResult_2[LDNode, str] = Decode_fromString(decoder_1, affiliation)
        if match_value.tag == 1:
            raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

        else: 
            return match_value.fields[0]


    except Exception as match_value_1:
        return LDOrganization.create(affiliation)



def PersonConversion_decomposeAffiliation_Z2F770004(affiliation: LDNode, context: LDContext | None=None) -> str:
    def predicate(n: str, affiliation: Any=affiliation, context: Any=context) -> bool:
        return n != LDOrganization.name()

    if is_empty(filter(predicate, affiliation.GetPropertyNames(context))):
        return LDOrganization.get_name_as_string(affiliation, context)

    else: 
        return to_string(0, encoder(affiliation))



def PersonConversion_composeAddress_Z721C83C5(address: str) -> Any:
    try: 
        def _arrow3924(__unit: None=None) -> LDNode:
            match_value: FSharpResult_2[LDNode, str] = Decode_fromString(decoder_1, address)
            if match_value.tag == 1:
                raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

            else: 
                return match_value.fields[0]


        return _arrow3924()

    except Exception as match_value_1:
        return address



def PersonConversion_decomposeAddress_4E60E31B(address: Any=None) -> str:
    if str(type(address)) == "<class \'str\'>":
        return address

    elif isinstance(address, LDNode):
        return to_string(0, encoder(address))

    else: 
        raise Exception("Address must be a string or a Json.LDNode")



def PersonConversion_composePerson_Z64D846DC(person: Person) -> LDNode:
    given_name: str
    match_value: str | None = person.FirstName
    if match_value is None:
        raise Exception("Person must have a given name")

    else: 
        given_name = match_value

    def f(term: OntologyAnnotation, person: Any=person) -> LDNode:
        return BaseTypes_composeDefinedTerm_ZDED3A0F(term)

    job_titles: Array[LDNode] | None = Option_fromSeq(ResizeArray_map(f, person.Roles))
    def f_1(c: Comment, person: Any=person) -> str:
        return to_string_1(c)

    disambiguating_descriptions: Array[str] | None = Option_fromSeq(ResizeArray_map(f_1, person.Comments))
    def mapping(address: str, person: Any=person) -> Any:
        return PersonConversion_composeAddress_Z721C83C5(address)

    address_1: Any | None = map(mapping, person.Address)
    def mapping_1(affiliation: str, person: Any=person) -> LDNode:
        return PersonConversion_composeAffiliation_Z721C83C5(affiliation)

    affiliation_1: LDNode | None = map(mapping_1, person.Affiliation)
    return LDPerson.create(given_name, person.ORCID, None, affiliation_1, person.EMail, person.LastName, None, job_titles, person.MidInitials, address_1, disambiguating_descriptions, person.Fax, person.Phone)


def PersonConversion_decomposePerson_Z6839B9E8(person: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Person:
    orcid: str | None = try_get_orcid_number(person.Id)
    address: str | None
    match_value: str | None = LDPerson.try_get_address_as_string(person, context)
    if match_value is None:
        match_value_1: LDNode | None = LDPerson.try_get_address_as_postal_address(person, graph, context)
        address = None if (match_value_1 is None) else PersonConversion_decomposeAddress_4E60E31B(match_value_1)

    else: 
        address = match_value

    def f(r: LDNode, person: Any=person, graph: Any=graph, context: Any=context) -> OntologyAnnotation:
        return BaseTypes_decomposeDefinedTerm_Z2F770004(r, context)

    roles: Array[OntologyAnnotation] = ResizeArray_map(f, LDPerson.get_job_titles_as_defined_term(person, graph, context))
    def f_1(s_1: str, person: Any=person, graph: Any=graph, context: Any=context) -> Comment:
        return Comment.from_string(s_1)

    comments: Array[Comment] = ResizeArray_map(f_1, LDPerson.get_disambiguating_descriptions_as_string(person, context))
    def mapping(a_3: LDNode, person: Any=person, graph: Any=graph, context: Any=context) -> str:
        return PersonConversion_decomposeAffiliation_Z2F770004(a_3, context)

    affiliation: str | None = map(mapping, LDPerson.try_get_affiliation(person, graph, context))
    return Person.create(orcid, LDPerson.try_get_family_name_as_string(person, context), LDPerson.get_given_name_as_string(person, context), LDPerson.try_get_additional_name_as_string(person, context), LDPerson.try_get_email_as_string(person, context), LDPerson.try_get_telephone_as_string(person, context), LDPerson.try_get_fax_number_as_string(person, context), address, affiliation, roles, comments)


__all__ = ["PersonConversion_reflection", "PersonConversion_get_orcidKey", "PersonConversion_composeAffiliation_Z721C83C5", "PersonConversion_decomposeAffiliation_Z2F770004", "PersonConversion_composeAddress_Z721C83C5", "PersonConversion_decomposeAddress_4E60E31B", "PersonConversion_composePerson_Z64D846DC", "PersonConversion_decomposePerson_Z6839B9E8"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from .py.main import Author, Constraint, Profile, License, CustomResourceDescriptor, Specification, Example, Guidance, UsedType, TextualResource, ResourceDescriptor, RootDataEntity


from __future__ import annotations
from typing import (Any, TypeVar)
from ...arctrl_core.Process.material import Material
from ...arctrl_core.Process.process import Process
from ...arctrl_core.Process.process_sequence import (get_samples, get_materials)
from ...arctrl_core.Process.sample import Sample
from ...fable_library.list import (FSharpList, choose, of_array)
from ...fable_library.option import map
from ...fable_library.seq import map as map_1
from ...fable_library.util import IEnumerable_1
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1)
from ..encode import try_include_list
from .material import ISAJson_encoder as ISAJson_encoder_1
from .sample import ISAJson_encoder

__A_ = TypeVar("__A_")

def encoder(id_map: Any | None, ps: FSharpList[Process]) -> IEncodable:
    samples: FSharpList[Sample] = get_samples(ps)
    materials: FSharpList[Material] = get_materials(ps)
    def chooser(tupled_arg: tuple[str, IEncodable | None], id_map: Any=id_map, ps: Any=ps) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2672(oa: Sample, id_map: Any=id_map, ps: Any=ps) -> IEncodable:
        return ISAJson_encoder(id_map, oa)

    def _arrow2673(c: Material, id_map: Any=id_map, ps: Any=ps) -> IEncodable:
        return ISAJson_encoder_1(id_map, c)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include_list("samples", _arrow2672, samples), try_include_list("otherMaterials", _arrow2673, materials)]))
    class ObjectExpr2674(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], id_map: Any=id_map, ps: Any=ps) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers.encode_object(arg)

    return ObjectExpr2674()


__all__ = ["encoder"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.Helper.collections_ import (ResizeArray_appendSingleton, ResizeArray_singleton)
from ...arctrl_core.Helper.identifier import create_missing_identifier
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Table.arc_table import ArcTable
from ...arctrl_core.Table.arc_table_aux import (Unchecked_alignByHeaders, ArcTableValues)
from ...arctrl_core.Table.arc_tables import ArcTables
from ...arctrl_core.Table.composite_cell import CompositeCell
from ...arctrl_core.Table.composite_header import CompositeHeader
from ...arctrl_file_system.file_system import FileSystem
from ...arctrl_rocrate.ldcontext import LDContext
from ...arctrl_rocrate.ldobject import (LDNode, LDGraph)
from ...arctrl_rocrate.LDTypes.lab_process import LDLabProcess
from ...arctrl_rocrate.LDTypes.lab_protocol import LDLabProtocol
from ...arctrl_rocrate.LDTypes.property_value import LDPropertyValue
from ...fable_library.list import (singleton, initialize, FSharpList, collect, of_seq, map as map_2)
from ...fable_library.option import map
from ...fable_library.range import range_big_int
from ...fable_library.seq import (fold, zip, to_list, delay, map as map_1)
from ...fable_library.seq2 import List_distinct
from ...fable_library.types import Array
from ...fable_library.util import (IEnumerable_1, get_enumerator, dispose, ignore, equals, safe_hash)
from .basic import (BaseTypes_composeDefinedTerm_ZDED3A0F, BaseTypes_composeComponent, BaseTypes_decomposeComponent_Z2F770004)
from .column_index import ARCtrl_ROCrate_LDNode__LDNode_TryGetColumnIndex
from .process import (ProcessConversion_tryGetProtocolType_Z6839B9E8, ProcessConversion_getProcessGetter, ProcessConversion_processToRows_Z6839B9E8, ProcessConversion_groupProcesses_Z27F0B586)

def CompositeRow_toProtocol(table_name: str, row: IEnumerable_1[tuple[CompositeHeader, CompositeCell]]) -> LDNode:
    def folder(p: LDNode, hc: tuple[CompositeHeader, CompositeCell], table_name: Any=table_name, row: Any=row) -> LDNode:
        (pattern_matching_result, oa, v, v_1, v_2, v_3) = (None, None, None, None, None, None)
        if hc[0].tag == 4:
            if hc[1].tag == 0:
                pattern_matching_result = 0
                oa = hc[1].fields[0]

            else: 
                pattern_matching_result = 6


        elif hc[0].tag == 7:
            if hc[1].tag == 1:
                pattern_matching_result = 1
                v = hc[1].fields[0]

            else: 
                pattern_matching_result = 6


        elif hc[0].tag == 6:
            if hc[1].tag == 1:
                pattern_matching_result = 2
                v_1 = hc[1].fields[0]

            else: 
                pattern_matching_result = 6


        elif hc[0].tag == 5:
            if hc[1].tag == 1:
                pattern_matching_result = 3
                v_2 = hc[1].fields[0]

            else: 
                pattern_matching_result = 6


        elif hc[0].tag == 8:
            if hc[1].tag == 1:
                pattern_matching_result = 4
                v_3 = hc[1].fields[0]

            else: 
                pattern_matching_result = 6


        elif hc[0].tag == 0:
            if hc[1].tag == 0:
                pattern_matching_result = 5

            elif hc[1].tag == 2:
                pattern_matching_result = 5

            else: 
                pattern_matching_result = 6


        else: 
            pattern_matching_result = 6

        if pattern_matching_result == 0:
            LDLabProtocol.set_intended_use_as_defined_term(p, BaseTypes_composeDefinedTerm_ZDED3A0F(oa))

        elif pattern_matching_result == 1:
            LDLabProtocol.set_version_as_string(p, v)

        elif pattern_matching_result == 2:
            LDLabProtocol.set_url(p, v_1)

        elif pattern_matching_result == 3:
            LDLabProtocol.set_description_as_string(p, v_2)

        elif pattern_matching_result == 4:
            LDLabProtocol.set_name_as_string(p, v_3)

        elif pattern_matching_result == 5:
            new_c: Array[LDNode] = ResizeArray_appendSingleton(BaseTypes_composeComponent(hc[0], hc[1]), LDLabProtocol.get_lab_equipments(p))
            LDLabProtocol.set_lab_equipments(p, new_c)

        return p

    return fold(folder, LDLabProtocol.create(table_name, table_name), row)


def ARCtrl_ArcTable__ArcTable_fromProtocol_Static_Z6839B9E8(p: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcTable:
    name: str = LDLabProtocol.get_name_as_string(p, context)
    t: ArcTable = ArcTable.init(name)
    enumerator: Any = get_enumerator(LDLabProtocol.get_components(p, graph, context))
    try: 
        while enumerator.System_Collections_IEnumerator_MoveNext():
            c: LDNode = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            pattern_input: tuple[CompositeHeader, CompositeCell] = BaseTypes_decomposeComponent_Z2F770004(c, context)
            t.AddColumn(pattern_input[0], ResizeArray_singleton(pattern_input[1]), ARCtrl_ROCrate_LDNode__LDNode_TryGetColumnIndex(c))

    finally: 
        dispose(enumerator)

    def mapping(d: str, p: Any=p, graph: Any=graph, context: Any=context) -> None:
        t.AddProtocolDescriptionColumn(ResizeArray_singleton(d))

    ignore(map(mapping, LDLabProtocol.try_get_description_as_string(p, context)))
    def mapping_1(d_1: str, p: Any=p, graph: Any=graph, context: Any=context) -> None:
        t.AddProtocolVersionColumn(ResizeArray_singleton(d_1))

    ignore(map(mapping_1, LDLabProtocol.try_get_version_as_string(p, context)))
    def mapping_2(d_2: OntologyAnnotation, p: Any=p, graph: Any=graph, context: Any=context) -> None:
        t.AddProtocolTypeColumn(ResizeArray_singleton(d_2))

    ignore(map(mapping_2, ProcessConversion_tryGetProtocolType_Z6839B9E8(p, None, context)))
    def mapping_3(d_3: str, p: Any=p, graph: Any=graph, context: Any=context) -> None:
        t.AddProtocolUriColumn(ResizeArray_singleton(d_3))

    ignore(map(mapping_3, LDLabProtocol.try_get_url(p, context)))
    t.AddProtocolNameColumn(ResizeArray_singleton(name))
    return t


def ARCtrl_ArcTable__ArcTable_GetProtocols(this: ArcTable) -> FSharpList[LDNode]:
    if this.RowCount == 0:
        def _arrow3878(__unit: None=None, this: Any=this) -> LDNode:
            source: Array[CompositeHeader] = this.Headers
            def folder(p: LDNode, h: CompositeHeader) -> LDNode:
                if h.tag == 0:
                    oa: OntologyAnnotation = h.fields[0]
                    match_value: str = oa.NameText
                    match_value_1: str = oa.TermAccessionOntobeeUrl
                    new_c: Array[LDNode] = ResizeArray_appendSingleton(LDPropertyValue.create_component(match_value, "Empty Component Value", None, match_value_1), LDLabProtocol.get_lab_equipments(p))
                    LDLabProtocol.set_lab_equipments(p, new_c)

                return p

            return fold(folder, LDLabProtocol.create(create_missing_identifier(), this.Name), source)

        return singleton(_arrow3878())

    else: 
        def _arrow3879(i: int, this: Any=this) -> LDNode:
            row: IEnumerable_1[tuple[CompositeHeader, CompositeCell]]
            source_2: Array[CompositeCell] = this.GetRow(i, True)
            row = zip(this.Headers, source_2)
            return CompositeRow_toProtocol(this.Name, row)

        class ObjectExpr3880:
            @property
            def Equals(self) -> Callable[[LDNode, LDNode], bool]:
                return equals

            @property
            def GetHashCode(self) -> Callable[[LDNode], int]:
                return safe_hash

        return List_distinct(initialize(this.RowCount, _arrow3879), ObjectExpr3880())



def ARCtrl_ArcTable__ArcTable_GetProcesses_5E660E5C(this: ArcTable, assay_name: str | None=None, study_name: str | None=None, fs: FileSystem | None=None) -> FSharpList[LDNode]:
    if this.RowCount == 0:
        return singleton(LDLabProcess.create(this.Name))

    else: 
        getter: Callable[[ArcTable, int], LDNode] = ProcessConversion_getProcessGetter(assay_name, study_name, this.Name, this.Headers, fs)
        def _arrow3882(__unit: None=None, this: Any=this, assay_name: Any=assay_name, study_name: Any=study_name, fs: Any=fs) -> IEnumerable_1[LDNode]:
            def _arrow3881(i: int) -> LDNode:
                return getter(this)(i)

            return map_1(_arrow3881, range_big_int(0, 1, this.RowCount - 1))

        return to_list(delay(_arrow3882))



def ARCtrl_ArcTable__ArcTable_fromProcesses_Static_Z3575FB5F(name: str, ps: FSharpList[LDNode], graph: LDGraph | None=None, context: LDContext | None=None) -> ArcTable:
    def mapping(p: LDNode, name: Any=name, ps: Any=ps, graph: Any=graph, context: Any=context) -> FSharpList[FSharpList[tuple[CompositeHeader, CompositeCell]]]:
        return of_seq(ProcessConversion_processToRows_Z6839B9E8(p, graph, context))

    tupled_arg: tuple[Array[CompositeHeader], ArcTableValues] = Unchecked_alignByHeaders(True, collect(mapping, ps))
    return ArcTable.from_arc_table_values(name, tupled_arg[0], tupled_arg[1])


def ARCtrl_ArcTables__ArcTables_GetProcesses_5E660E5C(this: ArcTables, assay_name: str | None=None, study_name: str | None=None, fs: FileSystem | None=None) -> FSharpList[LDNode]:
    def mapping(t: ArcTable, this: Any=this, assay_name: Any=assay_name, study_name: Any=study_name, fs: Any=fs) -> FSharpList[LDNode]:
        return ARCtrl_ArcTable__ArcTable_GetProcesses_5E660E5C(t, assay_name, study_name, fs)

    return collect(mapping, to_list(this.Tables))


def ARCtrl_ArcTables__ArcTables_fromProcesses_Static_Z27F0B586(ps: FSharpList[LDNode], graph: LDGraph | None=None, context: LDContext | None=None) -> ArcTables:
    def mapping_1(tupled_arg: tuple[str, FSharpList[LDNode]], ps: Any=ps, graph: Any=graph, context: Any=context) -> ArcTable:
        def mapping(p: LDNode, tupled_arg: Any=tupled_arg) -> FSharpList[FSharpList[tuple[CompositeHeader, CompositeCell]]]:
            return of_seq(ProcessConversion_processToRows_Z6839B9E8(p, graph, context))

        tupled_arg_1: tuple[Array[CompositeHeader], ArcTableValues] = Unchecked_alignByHeaders(True, collect(mapping, tupled_arg[1]))
        return ArcTable.from_arc_table_values(tupled_arg[0], tupled_arg_1[0], tupled_arg_1[1])

    return ArcTables(list(map_2(mapping_1, ProcessConversion_groupProcesses_Z27F0B586(ps, graph, context))))


__all__ = ["CompositeRow_toProtocol", "ARCtrl_ArcTable__ArcTable_fromProtocol_Static_Z6839B9E8", "ARCtrl_ArcTable__ArcTable_GetProtocols", "ARCtrl_ArcTable__ArcTable_GetProcesses_5E660E5C", "ARCtrl_ArcTable__ArcTable_fromProcesses_Static_Z3575FB5F", "ARCtrl_ArcTables__ArcTables_GetProcesses_5E660E5C", "ARCtrl_ArcTables__ArcTables_fromProcesses_Static_Z27F0B586"]




from __future__ import annotations
from dataclasses import dataclass
from typing import Any
from ..dynamic_obj.dynamic_obj import (DynamicObj, DynamicObj_reflection)
from ..dynamic_obj.dyn_obj import set_optional_property
from ..fable_library.option import value
from ..fable_library.reflection import (TypeInfo, string_type, option_type, record_type, class_type)
from ..fable_library.types import (Record, FSharpRef)
from .cwltypes import CWLType

def _expr486() -> TypeInfo:
    return record_type("ARCtrl.CWL.OutputBinding", [], OutputBinding, lambda: [("Glob", option_type(string_type))])


@dataclass(eq = False, repr = False, slots = True)
class OutputBinding(Record):
    Glob: str | None

OutputBinding_reflection = _expr486

def OutputBinding_create_6DFDD678(glob: str | None=None) -> OutputBinding:
    return OutputBinding(glob)


def _expr487() -> TypeInfo:
    return class_type("ARCtrl.CWL.CWLOutput", None, CWLOutput, DynamicObj_reflection())


class CWLOutput(DynamicObj):
    def __init__(self, name: str, type_: CWLType | None=None, output_binding: OutputBinding | None=None, output_source: str | None=None) -> None:
        super().__init__()
        this: FSharpRef[CWLOutput] = FSharpRef(None)
        self.name: str = name
        this.contents = self
        self.init_004013: int = 1
        set_optional_property("type", type_, this.contents)
        set_optional_property("outputBinding", output_binding, this.contents)
        set_optional_property("outputSource", output_source, this.contents)

    @property
    def Name(self, __unit: None=None) -> str:
        this: CWLOutput = self
        return this.name

    @property
    def Type_(self, __unit: None=None) -> CWLType | None:
        this: CWLOutput = self
        match_value: Any | None = this.TryGetPropertyValue("type")
        if match_value is not None:
            o: Any = value(match_value)
            return o if isinstance(o, CWLType) else None

        else: 
            return None


    @property
    def OutputBinding(self, __unit: None=None) -> OutputBinding | None:
        this: CWLOutput = self
        match_value: Any | None = this.TryGetPropertyValue("outputBinding")
        if match_value is not None:
            o: Any = value(match_value)
            return o if isinstance(o, OutputBinding) else None

        else: 
            return None


    @property
    def OutputSource(self, __unit: None=None) -> str | None:
        this: CWLOutput = self
        match_value: Any | None = this.TryGetPropertyValue("outputSource")
        if match_value is not None:
            o: Any = value(match_value)
            return o if (str(type(o)) == "<class \'str\'>") else None

        else: 
            return None



CWLOutput_reflection = _expr487

def CWLOutput__ctor_744035D(name: str, type_: CWLType | None=None, output_binding: OutputBinding | None=None, output_source: str | None=None) -> CWLOutput:
    return CWLOutput(name, type_, output_binding, output_source)


__all__ = ["OutputBinding_reflection", "OutputBinding_create_6DFDD678", "CWLOutput_reflection"]




from __future__ import annotations
from typing import (Any, TypeVar)
from ..arctrl_core.comment import Comment
from ..arctrl_core.ontology_source_reference import OntologySourceReference
from ..fable_library.list import (choose, of_array, FSharpList)
from ..fable_library.option import map
from ..fable_library.seq import map as map_1
from ..fable_library.string_ import replace
from ..fable_library.types import Array
from ..fable_library.util import IEnumerable_1
from ..thoth_json_core.decode import (object, IOptionalGetter, string, resize_array, IGetters)
from ..thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from .comment import (encoder as encoder_1, decoder as decoder_1, ISAJson_encoder as ISAJson_encoder_1)
from .context.rocrate.isa_ontology_source_reference_context import context_jsonvalue
from .decode import Decode_uri
from .encode import (try_include, try_include_seq)

__A_ = TypeVar("__A_")

def encoder(osr: OntologySourceReference) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], osr: Any=osr) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2113(value: str, osr: Any=osr) -> IEncodable:
        class ObjectExpr2112(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2112()

    def _arrow2115(value_2: str, osr: Any=osr) -> IEncodable:
        class ObjectExpr2114(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_2)

        return ObjectExpr2114()

    def _arrow2117(value_4: str, osr: Any=osr) -> IEncodable:
        class ObjectExpr2116(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_4)

        return ObjectExpr2116()

    def _arrow2119(value_6: str, osr: Any=osr) -> IEncodable:
        class ObjectExpr2118(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_6)

        return ObjectExpr2118()

    def _arrow2120(comment: Comment, osr: Any=osr) -> IEncodable:
        return encoder_1(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("description", _arrow2113, osr.Description), try_include("file", _arrow2115, osr.File), try_include("name", _arrow2117, osr.Name), try_include("version", _arrow2119, osr.Version), try_include_seq("comments", _arrow2120, osr.Comments)]))
    class ObjectExpr2121(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any], osr: Any=osr) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_4))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_4.encode_object(arg)

    return ObjectExpr2121()


def _arrow2127(get: IGetters) -> OntologySourceReference:
    def _arrow2122(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("description", Decode_uri)

    def _arrow2123(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("file", string)

    def _arrow2124(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("name", string)

    def _arrow2125(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("version", string)

    def _arrow2126(__unit: None=None) -> Array[Comment] | None:
        arg_9: Decoder_1[Array[Comment]] = resize_array(decoder_1)
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("comments", arg_9)

    return OntologySourceReference(_arrow2122(), _arrow2123(), _arrow2124(), _arrow2125(), _arrow2126())


decoder: Decoder_1[OntologySourceReference] = object(_arrow2127)

def ROCrate_genID(o: OntologySourceReference) -> str:
    match_value: str | None = o.File
    if match_value is None:
        match_value_1: str | None = o.Name
        if match_value_1 is None:
            return "#DummyOntologySourceRef"

        else: 
            return "#OntologySourceRef_" + replace(match_value_1, " ", "_")


    else: 
        return match_value



def ROCrate_encoder(osr: OntologySourceReference) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], osr: Any=osr) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2131(__unit: None=None, osr: Any=osr) -> IEncodable:
        value: str = ROCrate_genID(osr)
        class ObjectExpr2130(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2130()

    class ObjectExpr2132(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], osr: Any=osr) -> Any:
            return helpers_1.encode_string("OntologySourceReference")

    def _arrow2134(value_2: str, osr: Any=osr) -> IEncodable:
        class ObjectExpr2133(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_2)

        return ObjectExpr2133()

    def _arrow2136(value_4: str, osr: Any=osr) -> IEncodable:
        class ObjectExpr2135(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_4)

        return ObjectExpr2135()

    def _arrow2138(value_6: str, osr: Any=osr) -> IEncodable:
        class ObjectExpr2137(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_6)

        return ObjectExpr2137()

    def _arrow2140(value_8: str, osr: Any=osr) -> IEncodable:
        class ObjectExpr2139(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                return helpers_5.encode_string(value_8)

        return ObjectExpr2139()

    def _arrow2141(comment: Comment, osr: Any=osr) -> IEncodable:
        return encoder_1(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow2131()), ("@type", ObjectExpr2132()), try_include("description", _arrow2134, osr.Description), try_include("file", _arrow2136, osr.File), try_include("name", _arrow2138, osr.Name), try_include("version", _arrow2140, osr.Version), try_include_seq("comments", _arrow2141, osr.Comments), ("@context", context_jsonvalue)]))
    class ObjectExpr2142(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any], osr: Any=osr) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_6))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_6.encode_object(arg)

    return ObjectExpr2142()


def _arrow2148(get: IGetters) -> OntologySourceReference:
    def _arrow2143(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("description", Decode_uri)

    def _arrow2144(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("file", string)

    def _arrow2145(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("name", string)

    def _arrow2146(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("version", string)

    def _arrow2147(__unit: None=None) -> Array[Comment] | None:
        arg_9: Decoder_1[Array[Comment]] = resize_array(decoder_1)
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("comments", arg_9)

    return OntologySourceReference(_arrow2143(), _arrow2144(), _arrow2145(), _arrow2146(), _arrow2147())


ROCrate_decoder: Decoder_1[OntologySourceReference] = object(_arrow2148)

def ISAJson_encoder(id_map: Any | None, osr: OntologySourceReference) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], id_map: Any=id_map, osr: Any=osr) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2152(value: str, id_map: Any=id_map, osr: Any=osr) -> IEncodable:
        class ObjectExpr2151(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2151()

    def _arrow2154(value_2: str, id_map: Any=id_map, osr: Any=osr) -> IEncodable:
        class ObjectExpr2153(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_2)

        return ObjectExpr2153()

    def _arrow2156(value_4: str, id_map: Any=id_map, osr: Any=osr) -> IEncodable:
        class ObjectExpr2155(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_4)

        return ObjectExpr2155()

    def _arrow2158(value_6: str, id_map: Any=id_map, osr: Any=osr) -> IEncodable:
        class ObjectExpr2157(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_6)

        return ObjectExpr2157()

    def _arrow2159(comment: Comment, id_map: Any=id_map, osr: Any=osr) -> IEncodable:
        return ISAJson_encoder_1(id_map, comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("description", _arrow2152, osr.Description), try_include("file", _arrow2154, osr.File), try_include("name", _arrow2156, osr.Name), try_include("version", _arrow2158, osr.Version), try_include_seq("comments", _arrow2159, osr.Comments)]))
    class ObjectExpr2160(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any], id_map: Any=id_map, osr: Any=osr) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_4))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_4.encode_object(arg)

    return ObjectExpr2160()


ISAJson_decoder: Decoder_1[OntologySourceReference] = decoder

__all__ = ["encoder", "decoder", "ROCrate_genID", "ROCrate_encoder", "ROCrate_decoder", "ISAJson_encoder", "ISAJson_decoder"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ..fable_library.array_ import map as map_2
from ..fable_library.date import to_string
from ..fable_library.list import (is_empty as is_empty_1, map as map_3, FSharpList)
from ..fable_library.option import (map, default_arg)
from ..fable_library.seq import (is_empty, map as map_1, append, to_array)
from ..fable_library.string_ import (to_fail, printf)
from ..fable_library.types import Array
from ..fable_library.util import (to_enumerable, count, IDictionary)
from ..thoth_json_core.encode import (seq, list_1 as list_1_1, tuple2)
from ..thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Json)

_VALUE = TypeVar("_VALUE")

__A = TypeVar("__A")

__A_ = TypeVar("__A_")

_T = TypeVar("_T")

_KEY = TypeVar("_KEY")

def try_include(name: str, encoder: Callable[[_VALUE], IEncodable], value: Any | None=None) -> tuple[str, IEncodable | None]:
    return (name, map(encoder, value))


def try_include_seq(name: Any, encoder: Callable[[_VALUE], IEncodable], value: Any) -> tuple[__A, IEncodable | None]:
    return (name, None if is_empty(value) else seq(map_1(encoder, value)))


def try_include_array(name: Any, encoder: Callable[[_VALUE], IEncodable], value: Array[Any]) -> tuple[__A, IEncodable | None]:
    def _arrow1874(__unit: None=None, name: Any=name, encoder: Any=encoder, value: Any=value) -> IEncodable:
        values: Array[IEncodable] = map_2(encoder, value, None)
        class ObjectExpr1873(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[__A_]) -> __A_:
                def mapping(v: IEncodable) -> __A_:
                    return v.Encode(helpers)

                arg: Array[__A_] = map_2(mapping, values, None)
                return helpers.encode_array(arg)

        return ObjectExpr1873()

    return (name, None if (len(value) == 0) else _arrow1874())


def try_include_list(name: Any, encoder: Callable[[_VALUE], IEncodable], value: FSharpList[Any]) -> tuple[__A, IEncodable | None]:
    return (name, None if is_empty_1(value) else list_1_1(map_3(encoder, value)))


def try_include_list_opt(name: Any, encoder: Callable[[_VALUE], IEncodable], value: FSharpList[Any] | None=None) -> tuple[__A, IEncodable | None]:
    def _arrow1875(__unit: None=None, name: Any=name, encoder: Any=encoder, value: Any=value) -> IEncodable | None:
        o: FSharpList[_VALUE] = value
        return None if is_empty_1(o) else list_1_1(map_3(encoder, o))

    return (name, _arrow1875() if (value is not None) else None)


DefaultSpaces: int = 0

def default_spaces(spaces: int | None=None) -> int:
    return default_arg(spaces, DefaultSpaces)


def date_time(d: Any) -> IEncodable:
    value: str = to_string(d, "O", {}).split("+")[0]
    class ObjectExpr1876(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], d: Any=d) -> Any:
            return helpers.encode_string(value)

    return ObjectExpr1876()


def add_property_to_object(name: str, value: Json, obj: Json) -> Json:
    if obj.tag == 5:
        return Json(5, append(obj.fields[0], to_enumerable([(name, value)])))

    else: 
        raise Exception("Expected object")



def resize_array_or_singleton(encoder: Callable[[_T], IEncodable], values: Array[Any]) -> IEncodable:
    if len(values) == 1:
        return encoder(values[0])

    else: 
        return seq(map_1(encoder, values))



def dictionary(key_encoder: Callable[[_KEY], IEncodable], value_encoder: Callable[[_VALUE], IEncodable], values: IDictionary[Any, Any]) -> IEncodable:
    if count(values) == 0:
        class ObjectExpr1877(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any], key_encoder: Any=key_encoder, value_encoder: Any=value_encoder, values: Any=values) -> Any:
                return helpers.encode_null()

        return ObjectExpr1877()

    else: 
        def mapping(_arg: Any, key_encoder: Any=key_encoder, value_encoder: Any=value_encoder, values: Any=values) -> IEncodable:
            active_pattern_result: tuple[_KEY, _VALUE] = _arg
            return tuple2(key_encoder, value_encoder, active_pattern_result[0], active_pattern_result[1])

        return seq(to_array(map_1(mapping, values)))



def int_dictionary(value_encoder: Callable[[_VALUE], IEncodable], values: IDictionary[int, Any]) -> IEncodable:
    if count(values) == 0:
        class ObjectExpr1878(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any], value_encoder: Any=value_encoder, values: Any=values) -> Any:
                return helpers.encode_null()

        return ObjectExpr1878()

    else: 
        def mapping(_arg: Any, value_encoder: Any=value_encoder, values: Any=values) -> IEncodable:
            active_pattern_result: tuple[int, _VALUE] = _arg
            k: int = active_pattern_result[0] or 0
            if True if (k > 2147483647) else (k < -2147483648):
                to_fail(printf("Key %d is out of bounds for Int32"))(k)

            def _arrow1880(value: int, _arg: Any=_arg) -> IEncodable:
                class ObjectExpr1879(IEncodable):
                    def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                        return helpers_1.encode_signed_integral_number(value)

                return ObjectExpr1879()

            return tuple2(_arrow1880, value_encoder, k, active_pattern_result[1])

        return seq(to_array(map_1(mapping, values)))



__all__ = ["try_include", "try_include_seq", "try_include_array", "try_include_list", "try_include_list_opt", "DefaultSpaces", "default_spaces", "date_time", "add_property_to_object", "resize_array_or_singleton", "dictionary", "int_dictionary"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...fable_library.list import (empty, FSharpList, map, append)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.string_ import (to_text, printf)
from .expression import (RequiredSource_1__ctor_2B595, RequiredSource_1, OptionalSource_1__ctor_2B595, OptionalSource_1, RequiredSource_1__get_Source, OptionalSource_1__get_Source)
from .types import (SheetEntity_1_some_2B595, SheetEntity_1, ColumnElement, Message__MapText_11D407F6, Message, SheetEntity_1__get_Messages)

def _expr314() -> TypeInfo:
    return class_type("FsSpreadsheet.DSL.ColumnBuilder", None, ColumnBuilder)


class ColumnBuilder:
    def __init__(self, __unit: None=None) -> None:
        pass


ColumnBuilder_reflection = _expr314

def ColumnBuilder__ctor(__unit: None=None) -> ColumnBuilder:
    return ColumnBuilder(__unit)


def ColumnBuilder_get_Empty(__unit: None=None) -> SheetEntity_1[FSharpList[ColumnElement]]:
    return SheetEntity_1_some_2B595(empty())


def ColumnBuilder__SignMessages_3F89FFFC(this: ColumnBuilder, messages: FSharpList[Message]) -> FSharpList[Message]:
    def mapping(m: Message, this: Any=this, messages: Any=messages) -> Message:
        def _arrow315(__unit: None=None, m: Any=m) -> Callable[[str], str]:
            clo: Callable[[str], str] = to_text(printf("In Column: %s"))
            return clo

        return Message__MapText_11D407F6(m, _arrow315())

    return map(mapping, messages)


def ColumnBuilder__Combine_Z14A04580(this: ColumnBuilder, wx1: SheetEntity_1[FSharpList[ColumnElement]], wx2: SheetEntity_1[FSharpList[ColumnElement]]) -> SheetEntity_1[FSharpList[ColumnElement]]:
    (pattern_matching_result, l1, l2, messages1, messages2, messages2_1, messages1_1, f1, messages1_2, messages2_2, f2, messages1_3, messages2_3, messages1_4, messages2_4) = (None, None, None, None, None, None, None, None, None, None, None, None, None, None, None)
    if wx1.tag == 2:
        if wx2.tag == 2:
            pattern_matching_result = 1
            messages2_1 = wx2.fields[0]

        else: 
            pattern_matching_result = 2
            messages1_1 = wx1.fields[0]


    elif wx1.tag == 1:
        if wx2.tag == 0:
            pattern_matching_result = 4
            f2 = wx2.fields[0]
            messages1_3 = wx1.fields[0]
            messages2_3 = wx2.fields[1]

        elif wx2.tag == 1:
            pattern_matching_result = 5
            messages1_4 = wx1.fields[0]
            messages2_4 = wx2.fields[0]

        else: 
            pattern_matching_result = 1
            messages2_1 = wx2.fields[0]


    elif wx2.tag == 2:
        pattern_matching_result = 1
        messages2_1 = wx2.fields[0]

    elif wx2.tag == 1:
        pattern_matching_result = 3
        f1 = wx1.fields[0]
        messages1_2 = wx1.fields[1]
        messages2_2 = wx2.fields[0]

    else: 
        pattern_matching_result = 0
        l1 = wx1.fields[0]
        l2 = wx2.fields[0]
        messages1 = wx1.fields[1]
        messages2 = wx2.fields[1]

    if pattern_matching_result == 0:
        return SheetEntity_1(0, append(l1, l2), append(messages1, messages2))

    elif pattern_matching_result == 1:
        return SheetEntity_1(2, append(SheetEntity_1__get_Messages(wx1), messages2_1))

    elif pattern_matching_result == 2:
        return SheetEntity_1(2, append(messages1_1, SheetEntity_1__get_Messages(wx2)))

    elif pattern_matching_result == 3:
        return SheetEntity_1(0, f1, append(messages1_2, messages2_2))

    elif pattern_matching_result == 4:
        return SheetEntity_1(0, f2, append(messages1_3, messages2_3))

    elif pattern_matching_result == 5:
        return SheetEntity_1(1, append(messages1_4, messages2_4))



def ColumnBuilder__Combine_89C473F(this: ColumnBuilder, wx1: RequiredSource_1[None], wx2: SheetEntity_1[FSharpList[ColumnElement]]) -> RequiredSource_1[SheetEntity_1[FSharpList[ColumnElement]]]:
    return RequiredSource_1__ctor_2B595(wx2)


def ColumnBuilder__Combine_740C071F(this: ColumnBuilder, wx1: SheetEntity_1[FSharpList[ColumnElement]], wx2: RequiredSource_1[None]) -> RequiredSource_1[SheetEntity_1[FSharpList[ColumnElement]]]:
    return RequiredSource_1__ctor_2B595(wx1)


def ColumnBuilder__Combine_1B6D0B58(this: ColumnBuilder, wx1: OptionalSource_1[None], wx2: SheetEntity_1[FSharpList[ColumnElement]]) -> OptionalSource_1[SheetEntity_1[FSharpList[ColumnElement]]]:
    return OptionalSource_1__ctor_2B595(wx2)


def ColumnBuilder__Combine_Z4E78BCA8(this: ColumnBuilder, wx1: SheetEntity_1[FSharpList[ColumnElement]], wx2: OptionalSource_1[None]) -> OptionalSource_1[SheetEntity_1[FSharpList[ColumnElement]]]:
    return OptionalSource_1__ctor_2B595(wx1)


def ColumnBuilder__Combine_F0E07EF(this: ColumnBuilder, wx1: RequiredSource_1[SheetEntity_1[FSharpList[ColumnElement]]], wx2: SheetEntity_1[FSharpList[ColumnElement]]) -> RequiredSource_1[SheetEntity_1[FSharpList[ColumnElement]]]:
    return RequiredSource_1__ctor_2B595(ColumnBuilder__Combine_Z14A04580(this, RequiredSource_1__get_Source(wx1), wx2))


def ColumnBuilder__Combine_48CBECF(this: ColumnBuilder, wx1: SheetEntity_1[FSharpList[ColumnElement]], wx2: RequiredSource_1[SheetEntity_1[FSharpList[ColumnElement]]]) -> RequiredSource_1[SheetEntity_1[FSharpList[ColumnElement]]]:
    return RequiredSource_1__ctor_2B595(ColumnBuilder__Combine_Z14A04580(this, wx1, RequiredSource_1__get_Source(wx2)))


def ColumnBuilder__Combine_Z172253D8(this: ColumnBuilder, wx1: OptionalSource_1[SheetEntity_1[FSharpList[ColumnElement]]], wx2: SheetEntity_1[FSharpList[ColumnElement]]) -> OptionalSource_1[SheetEntity_1[FSharpList[ColumnElement]]]:
    return OptionalSource_1__ctor_2B595(ColumnBuilder__Combine_Z14A04580(this, OptionalSource_1__get_Source(wx1), wx2))


def ColumnBuilder__Combine_Z73B7C9D8(this: ColumnBuilder, wx1: SheetEntity_1[FSharpList[ColumnElement]], wx2: OptionalSource_1[SheetEntity_1[FSharpList[ColumnElement]]]) -> OptionalSource_1[SheetEntity_1[FSharpList[ColumnElement]]]:
    return OptionalSource_1__ctor_2B595(ColumnBuilder__Combine_Z14A04580(this, wx1, OptionalSource_1__get_Source(wx2)))


__all__ = ["ColumnBuilder_reflection", "ColumnBuilder_get_Empty", "ColumnBuilder__SignMessages_3F89FFFC", "ColumnBuilder__Combine_Z14A04580", "ColumnBuilder__Combine_89C473F", "ColumnBuilder__Combine_740C071F", "ColumnBuilder__Combine_1B6D0B58", "ColumnBuilder__Combine_Z4E78BCA8", "ColumnBuilder__Combine_F0E07EF", "ColumnBuilder__Combine_48CBECF", "ColumnBuilder__Combine_Z172253D8", "ColumnBuilder__Combine_Z73B7C9D8"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..arctrl_core.arc_types import ArcRun
from ..arctrl_core.comment import Remark
from ..arctrl_core.data_map import DataMap
from ..arctrl_core.Helper.identifier import create_missing_identifier
from ..arctrl_core.person import Person
from ..arctrl_core.Table.arc_table import ArcTable
from ..fable_library.list import (FSharpList, empty, of_seq)
from ..fable_library.seq import (delay, append, singleton, iterate_indexed, map, try_find, choose, try_pick, is_empty)
from ..fable_library.string_ import (to_fail, printf, to_console)
from ..fable_library.types import Array
from ..fable_library.util import (get_enumerator, IEnumerable_1, IEnumerator, to_enumerable)
from ..fs_spreadsheet.fs_row import FsRow
from ..fs_spreadsheet.fs_workbook import FsWorkbook
from ..fs_spreadsheet.fs_worksheet import FsWorksheet
from .AnnotationTable.arc_table import (try_from_fs_worksheet, to_fs_worksheet)
from .DataMapTable.data_map_table import try_from_fs_worksheet as try_from_fs_worksheet_1
from .Metadata.contacts import (from_rows as from_rows_1, to_rows as to_rows_1)
from .Metadata.run import (from_rows, to_rows)
from .Metadata.sparse_table import (SparseRowModule_tryGetValueAt, SparseRowModule_fromValues, SparseRowModule_writeToSheet, SparseRowModule_fromFsRow, SparseRowModule_getAllValues, SparseRowModule_fromAllValues)

def ArcRun_fromRows(rows: IEnumerable_1[IEnumerable_1[tuple[int, str]]]) -> ArcRun:
    en: IEnumerator[IEnumerable_1[tuple[int, str]]] = get_enumerator(rows)
    def loop(last_row_mut: str | None, run_mut: ArcRun | None, performers_mut: FSharpList[Person], row_number_mut: int, rows: Any=rows) -> ArcRun:
        while True:
            (last_row, run, performers, row_number) = (last_row_mut, run_mut, performers_mut, row_number_mut)
            (pattern_matching_result,) = (None,)
            if last_row is not None:
                if last_row == "RUN":
                    pattern_matching_result = 0

                elif last_row == "RUN PERFORMERS":
                    pattern_matching_result = 1

                else: 
                    pattern_matching_result = 2


            else: 
                pattern_matching_result = 2

            if pattern_matching_result == 0:
                pattern_input: tuple[str | None, int, FSharpList[Remark], ArcRun] = from_rows(row_number + 1, en)
                last_row_mut = pattern_input[0]
                run_mut = pattern_input[3]
                performers_mut = performers
                row_number_mut = pattern_input[1]
                continue

            elif pattern_matching_result == 1:
                pattern_input_1: tuple[str | None, int, FSharpList[Remark], FSharpList[Person]] = from_rows_1("Run Person", row_number + 1, en)
                last_row_mut = pattern_input_1[0]
                run_mut = run
                performers_mut = pattern_input_1[3]
                row_number_mut = pattern_input_1[1]
                continue

            elif pattern_matching_result == 2:
                if run is not None:
                    run_2: ArcRun = run
                    run_2.Performers = list(performers)
                    return run_2

                else: 
                    return ArcRun.create(create_missing_identifier(), None, None, None, None, None, None, None, None, list(performers))


            break

    if en.System_Collections_IEnumerator_MoveNext():
        return loop(SparseRowModule_tryGetValueAt(0, en.System_Collections_Generic_IEnumerator_1_get_Current()), None, empty(), 1)

    else: 
        raise Exception("empty run metadata sheet")



def ArcRun_toRows(run: ArcRun) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    def _arrow3648(__unit: None=None, run: Any=run) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
        def _arrow3647(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
            def _arrow3646(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                def _arrow3645(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                    return to_rows_1("Run Person", of_seq(run.Performers))

                return append(singleton(SparseRowModule_fromValues(to_enumerable(["RUN PERFORMERS"]))), delay(_arrow3645))

            return append(to_rows(run), delay(_arrow3646))

        return append(singleton(SparseRowModule_fromValues(to_enumerable(["RUN"]))), delay(_arrow3647))

    return delay(_arrow3648)


def ArcRun_toMetadataSheet(run: ArcRun) -> FsWorksheet:
    sheet: FsWorksheet = FsWorksheet("isa_run")
    def action(row_i: int, r: IEnumerable_1[tuple[int, str]], run: Any=run) -> None:
        SparseRowModule_writeToSheet(row_i + 1, r, sheet)

    iterate_indexed(action, ArcRun_toRows(run))
    return sheet


def ArcRun_fromMetadataSheet(sheet: FsWorksheet) -> ArcRun:
    try: 
        return ArcRun_fromRows(map(SparseRowModule_fromFsRow, sheet.Rows))

    except Exception as err:
        arg: str = str(err)
        return to_fail(printf("Failed while parsing metadatasheet: %s"))(arg)



def ArcRun_toMetadataCollection(run: ArcRun) -> IEnumerable_1[IEnumerable_1[str | None]]:
    def mapping(row: IEnumerable_1[tuple[int, str]], run: Any=run) -> IEnumerable_1[str | None]:
        return SparseRowModule_getAllValues(row)

    return map(mapping, ArcRun_toRows(run))


def ArcRun_fromMetadataCollection(collection: IEnumerable_1[IEnumerable_1[str | None]]) -> ArcRun:
    try: 
        return ArcRun_fromRows(map(SparseRowModule_fromAllValues, collection))

    except Exception as err:
        arg: str = str(err)
        return to_fail(printf("Failed while parsing metadatasheet: %s"))(arg)



def ArcRun_isMetadataSheetName(name: str) -> bool:
    return name == "isa_run"


def ArcRun_isMetadataSheet(sheet: FsWorksheet) -> bool:
    return ArcRun_isMetadataSheetName(sheet.Name)


def ArcRun_tryGetMetadataSheet(doc: FsWorkbook) -> FsWorksheet | None:
    def predicate(sheet: FsWorksheet, doc: Any=doc) -> bool:
        return ArcRun_isMetadataSheet(sheet)

    return try_find(predicate, doc.GetWorksheets())


def ARCtrl_ArcRun__ArcRun_fromFsWorkbook_Static_32154C9D(doc: FsWorkbook) -> ArcRun:
    try: 
        run_metadata: ArcRun
        match_value: FsWorksheet | None = ArcRun_tryGetMetadataSheet(doc)
        if match_value is None:
            to_console(printf("Cannot retrieve metadata: Run file does not contain \"%s\" sheet."))("isa_run")
            run_metadata = ArcRun.create(create_missing_identifier())

        else: 
            run_metadata = ArcRun_fromMetadataSheet(match_value)

        sheets: Array[FsWorksheet] = doc.GetWorksheets()
        annotation_tables: IEnumerable_1[ArcTable] = choose(try_from_fs_worksheet, sheets)
        datamap_sheet: DataMap | None = try_pick(try_from_fs_worksheet_1, sheets)
        run_metadata.DataMap = datamap_sheet
        if not is_empty(annotation_tables):
            run_metadata.Tables = list(annotation_tables)

        return run_metadata

    except Exception as err:
        arg_1: str = str(err)
        return to_fail(printf("Could not parse assay: \n%s"))(arg_1)



def ARCtrl_ArcRun__ArcRun_toFsWorkbook_Static_Z3EFAF6F8(run: ArcRun) -> FsWorkbook:
    doc: FsWorkbook = FsWorkbook()
    metadata_sheet: FsWorksheet = ArcRun_toMetadataSheet(run)
    doc.AddWorksheet(metadata_sheet)
    def action(i: int, arg: ArcTable, run: Any=run) -> None:
        sheet: FsWorksheet = to_fs_worksheet(i, arg)
        doc.AddWorksheet(sheet)

    iterate_indexed(action, run.Tables)
    return doc


def ARCtrl_ArcRun__ArcRun_ToFsWorkbook(this: ArcRun) -> FsWorkbook:
    return ARCtrl_ArcRun__ArcRun_toFsWorkbook_Static_Z3EFAF6F8(this)


__all__ = ["ArcRun_fromRows", "ArcRun_toRows", "ArcRun_toMetadataSheet", "ArcRun_fromMetadataSheet", "ArcRun_toMetadataCollection", "ArcRun_fromMetadataCollection", "ArcRun_isMetadataSheetName", "ArcRun_isMetadataSheet", "ArcRun_tryGetMetadataSheet", "ARCtrl_ArcRun__ArcRun_fromFsWorkbook_Static_32154C9D", "ARCtrl_ArcRun__ArcRun_toFsWorkbook_Static_Z3EFAF6F8", "ARCtrl_ArcRun__ArcRun_ToFsWorkbook"]




from __future__ import annotations
from typing import Any
from ...arctrl_core.Table.composite_header import IOType
from ...fable_library.types import to_string
from ...thoth_json_core.decode import (and_then, succeed, string)
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)

def encoder(io: IOType) -> IEncodable:
    value: str = to_string(io)
    class ObjectExpr2765(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], io: Any=io) -> Any:
            return helpers.encode_string(value)

    return ObjectExpr2765()


def cb(s: str) -> Decoder_1[IOType]:
    return succeed(IOType.of_string(s))


decoder: Decoder_1[IOType] = and_then(cb, string)

__all__ = ["encoder", "decoder"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...fable_library.array_ import contains
from ...fable_library.list import (is_empty as is_empty_1, tail, FSharpList, head, append, singleton, empty)
from ...fable_library.map_util import (add_to_dict, try_get_value, remove_from_dict)
from ...fable_library.mutable_map import Dictionary
from ...fable_library.option import (some, value as value_1)
from ...fable_library.range import range_big_int
from ...fable_library.seq import (is_empty, iterate, map)
from ...fable_library.seq2 import group_by
from ...fable_library.types import (FSharpRef, Array)
from ...fable_library.util import (equals, structural_hash, IEnumerable_1, ignore, get_enumerator, dispose, min, compare_primitives)

_T = TypeVar("_T")

_U = TypeVar("_U")

_KEY_ = TypeVar("_KEY_")

_KEY = TypeVar("_KEY")

__B = TypeVar("__B")

__A = TypeVar("__A")

__C = TypeVar("__C")

__A_ = TypeVar("__A_")

_A = TypeVar("_A")

_A_ = TypeVar("_A_")

def Option_fromValueWithDefault(d: Any, v: Any) -> Any | None:
    if equals(d, v):
        return None

    else: 
        return some(v)



def Option_mapDefault(d: Any, f: Callable[[_T], _T], o: Any | None=None) -> Any | None:
    return Option_fromValueWithDefault(d, f(d) if (o is None) else f(value_1(o)))


def Option_mapOrDefault(d: Any | None, f: Callable[[_U], _T], o: Any | None=None) -> Any | None:
    if o is None:
        return d

    else: 
        return some(f(value_1(o)))



def Option_fromSeq(v: Any | None=None) -> Any | None:
    if is_empty(v):
        return None

    else: 
        return some(v)



def List_tryPickAndRemove(f: Callable[[_T], _U | None], lst: FSharpList[Any]) -> tuple[_U | None, FSharpList[_T]]:
    def loop(new_list_mut: FSharpList[_T], remaining_list_mut: FSharpList[_T], f: Any=f, lst: Any=lst) -> tuple[_U | None, FSharpList[_T]]:
        while True:
            (new_list, remaining_list) = (new_list_mut, remaining_list_mut)
            if not is_empty_1(remaining_list):
                t: FSharpList[_T] = tail(remaining_list)
                h: _T = head(remaining_list)
                match_value: _U | None = f(h)
                if match_value is None:
                    new_list_mut = append(new_list, singleton(h))
                    remaining_list_mut = t
                    continue

                else: 
                    return (some(value_1(match_value)), append(new_list, t))


            else: 
                return (None, new_list)

            break

    return loop(empty(), lst)


def Dictionary_addOrUpdate(key: Any, value: Any, dict_1: Any) -> None:
    if key in dict_1:
        dict_1[key] = value

    else: 
        add_to_dict(dict_1, key, value)



def Dictionary_ofSeq(s: IEnumerable_1[tuple[_KEY, _T]]) -> Any:
    class ObjectExpr637:
        @property
        def Equals(self) -> Callable[[_KEY_, _KEY_], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[_KEY_], int]:
            return structural_hash

    dict_1: Any = Dictionary([], ObjectExpr637())
    def action(tupled_arg: tuple[_KEY, _T], s: Any=s) -> None:
        add_to_dict(dict_1, tupled_arg[0], tupled_arg[1])

    iterate(action, s)
    return dict_1


def Dictionary_tryFind(key: Any, dict_1: Any) -> Any | None:
    pattern_input: tuple[bool, _T]
    out_arg: _T = None
    def _arrow638(__unit: None=None, key: Any=key, dict_1: Any=dict_1) -> _T:
        return out_arg

    def _arrow639(v: _T | None=None, key: Any=key, dict_1: Any=dict_1) -> None:
        nonlocal out_arg
        out_arg = v

    pattern_input = (try_get_value(dict_1, key, FSharpRef(_arrow638, _arrow639)), out_arg)
    if pattern_input[0]:
        return some(pattern_input[1])

    else: 
        return None



def Dictionary_ofSeqWithMerge(merge: Callable[[_T, _T], _T], s: IEnumerable_1[tuple[_KEY, _T]]) -> Any:
    class ObjectExpr640:
        @property
        def Equals(self) -> Callable[[_KEY_, _KEY_], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[_KEY_], int]:
            return structural_hash

    dict_1: Any = Dictionary([], ObjectExpr640())
    def action(tupled_arg: tuple[_KEY, _T], merge: Any=merge, s: Any=s) -> None:
        k: _KEY = tupled_arg[0]
        v: _T = tupled_arg[1]
        match_value: _T | None = Dictionary_tryFind(k, dict_1)
        if match_value is None:
            add_to_dict(dict_1, k, v)

        else: 
            v_0027: _T = value_1(match_value)
            ignore(remove_from_dict(dict_1, k))
            add_to_dict(dict_1, k, merge(v_0027, v))


    iterate(action, s)
    return dict_1


def Dictionary_init(__unit: None=None) -> Any:
    return dict()


def Dictionary_items(dict_1: Any) -> IEnumerable_1[Any]:
    return dict_1.items()


def StringDictionary_ofSeq(s: IEnumerable_1[tuple[str, str]]) -> Any:
    return dict(s)


def IntDictionary_ofSeq(s: IEnumerable_1[tuple[int, _T]]) -> Any:
    return dict(s)


def ResizeArray_create(i: int, v: Any) -> Array[Any]:
    a: Array[_T] = []
    if i > 0:
        for for_loop_var in range(1, i + 1, 1):
            (a.append(v))

    return a


def ResizeArray_singleton(a: Any | None=None) -> Array[Any]:
    b: Array[_T] = []
    (b.append(a))
    return b


def ResizeArray_map(f: Callable[[__A], __B], a: Array[Any]) -> Array[Any]:
    b: Array[__B] = []
    enumerator: Any = get_enumerator(a)
    try: 
        while enumerator.System_Collections_IEnumerator_MoveNext():
            i: __A = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            (b.append(f(i)))

    finally: 
        dispose(enumerator)

    return b


def ResizeArray_mapi(f: Callable[[int, __A], __B], a: Array[Any]) -> Array[Any]:
    b: Array[__B] = []
    for i in range(0, (len(a) - 1) + 1, 1):
        (b.append(f(i, a[i])))
    return b


def ResizeArray_map2(f: Callable[[__A, __B], __C], a: Array[Any], b: Array[Any]) -> Array[Any]:
    c: Array[__C] = []
    def _arrow641(x: int, y: int, f: Any=f, a: Any=a, b: Any=b) -> int:
        return compare_primitives(x, y)

    n: int = min(_arrow641, len(a), len(b)) or 0
    for i in range(0, (n - 1) + 1, 1):
        (c.append(f(a[i], b[i])))
    return c


def ResizeArray_choose(f: Callable[[__A], __B | None], a: Array[Any]) -> Array[Any]:
    b: Array[__B] = []
    enumerator: Any = get_enumerator(a)
    try: 
        while enumerator.System_Collections_IEnumerator_MoveNext():
            match_value: __B | None = f(enumerator.System_Collections_Generic_IEnumerator_1_get_Current())
            if match_value is None:
                pass

            else: 
                x: __B = value_1(match_value)
                (b.append(x))


    finally: 
        dispose(enumerator)

    return b


def ResizeArray_filter(f: Callable[[__A], bool], a: Array[Any]) -> Array[Any]:
    b: Array[__A] = []
    enumerator: Any = get_enumerator(a)
    try: 
        while enumerator.System_Collections_IEnumerator_MoveNext():
            i: __A = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            if f(i):
                (b.append(i))


    finally: 
        dispose(enumerator)

    return b


def ResizeArray_fold(f: Callable[[__A, __B], __A], s: Any, a: Array[Any]) -> Any:
    state: __A = s
    enumerator: Any = get_enumerator(a)
    try: 
        while enumerator.System_Collections_IEnumerator_MoveNext():
            i: __B = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            state = f(state, i)

    finally: 
        dispose(enumerator)

    return state


def ResizeArray_foldBack(f: Callable[[__A, __B], __B], a: Array[Any], s: Any) -> Any:
    state: __B = s
    enumerator: Any = get_enumerator(a)
    try: 
        while enumerator.System_Collections_IEnumerator_MoveNext():
            i: __A = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            state = f(i, state)

    finally: 
        dispose(enumerator)

    return state


def ResizeArray_iter(f: Callable[[__A], None], a: Array[Any]) -> None:
    enumerator: Any = get_enumerator(a)
    try: 
        while enumerator.System_Collections_IEnumerator_MoveNext():
            f(enumerator.System_Collections_Generic_IEnumerator_1_get_Current())

    finally: 
        dispose(enumerator)



def ResizeArray_iteri(f: Callable[[int, __A], None], a: Array[Any]) -> None:
    for i in range(0, (len(a) - 1) + 1, 1):
        f(i, a[i])


def ResizeArray_init(n: int, f: Callable[[int], _T]) -> Array[Any]:
    a: Array[_T] = []
    for i in range(0, (n - 1) + 1, 1):
        (a.append(f(i)))
    return a


def ResizeArray_reduce(f: Callable[[__A, __A], __A], a: Array[Any]) -> Any:
    if len(a) == 0:
        raise Exception("ResizeArray.reduce: empty array")

    elif len(a) == 1:
        return a[0]

    else: 
        a_5: Array[__A] = a
        state: __A = a_5[0]
        for i in range(1, (len(a_5) - 1) + 1, 1):
            state = f(state, a_5[i])
        return state



def ResizeArray_collect(f: Callable[[__A], __B], a: Array[Any]) -> Array[Any]:
    b: Array[__C] = []
    enumerator: Any = get_enumerator(a)
    try: 
        while enumerator.System_Collections_IEnumerator_MoveNext():
            with get_enumerator(f(enumerator.System_Collections_Generic_IEnumerator_1_get_Current())) as enumerator_1:
                while enumerator_1.System_Collections_IEnumerator_MoveNext():
                    j: __C = enumerator_1.System_Collections_Generic_IEnumerator_1_get_Current()
                    (b.append(j))

    finally: 
        dispose(enumerator)

    return b


def ResizeArray_distinct(a: Array[Any]) -> Array[Any]:
    b: Array[__A] = []
    enumerator: Any = get_enumerator(a)
    try: 
        while enumerator.System_Collections_IEnumerator_MoveNext():
            i: __A = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            class ObjectExpr642:
                @property
                def Equals(self) -> Callable[[__A_, __A_], bool]:
                    return equals

                @property
                def GetHashCode(self) -> Callable[[__A_], int]:
                    return structural_hash

            if not contains(i, b, ObjectExpr642()):
                (b.append(i))


    finally: 
        dispose(enumerator)

    return b


def ResizeArray_isEmpty(a: Array[Any]) -> bool:
    return len(a) == 0


def ResizeArray_append(a: Array[Any], b: Array[Any]) -> Array[Any]:
    c: Array[__A] = []
    enumerator: Any = get_enumerator(a)
    try: 
        while enumerator.System_Collections_IEnumerator_MoveNext():
            i: __A = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            (c.append(i))

    finally: 
        dispose(enumerator)

    enumerator_1: Any = get_enumerator(b)
    try: 
        while enumerator_1.System_Collections_IEnumerator_MoveNext():
            i_1: __A = enumerator_1.System_Collections_Generic_IEnumerator_1_get_Current()
            (c.append(i_1))

    finally: 
        dispose(enumerator_1)

    return c


def ResizeArray_appendSingleton(b: Any, a: Array[Any]) -> Array[Any]:
    c: Array[_T] = []
    enumerator: Any = get_enumerator(a)
    try: 
        while enumerator.System_Collections_IEnumerator_MoveNext():
            i: _T = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            (c.append(i))

    finally: 
        dispose(enumerator)

    (c.append(b))
    return c


def ResizeArray_indexed(a: Array[Any]) -> Array[tuple[int, __A]]:
    b: Array[tuple[int, __A]] = []
    for i in range(0, (len(a) - 1) + 1, 1):
        (b.append((i, a[i])))
    return b


def ResizeArray_rev(a: Array[Any]) -> Array[Any]:
    b: Array[__A] = []
    with get_enumerator(range_big_int(len(a) - 1, -1, 0)) as enumerator:
        while enumerator.System_Collections_IEnumerator_MoveNext():
            i: int = enumerator.System_Collections_Generic_IEnumerator_1_get_Current() or 0
            (b.append(a[i]))
    return b


def ResizeArray_take(n: int, a: Array[Any]) -> Array[Any]:
    b: Array[__A] = []
    def _arrow643(x: int, y: int, n: Any=n, a: Any=a) -> int:
        return compare_primitives(x, y)

    n_1: int = min(_arrow643, n, len(a)) or 0
    for i in range(0, (n_1 - 1) + 1, 1):
        (b.append(a[i]))
    return b


def ResizeArray_groupBy(f: Callable[[_T], _A], a: Array[Any]) -> Array[tuple[_A, Array[_T]]]:
    def mapping(tupled_arg: tuple[_A, IEnumerable_1[_T]], f: Any=f, a: Any=a) -> tuple[_A, Array[_T]]:
        return (tupled_arg[0], list(tupled_arg[1]))

    class ObjectExpr644:
        @property
        def Equals(self) -> Callable[[_A_, _A_], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[_A_], int]:
            return structural_hash

    return list(map(mapping, group_by(f, a, ObjectExpr644())))


def ResizeArray_tryPick(f: Callable[[_T], __A | None], a: Array[Any]) -> Any | None:
    def loop(i_mut: int, f: Any=f, a: Any=a) -> __A | None:
        while True:
            (i,) = (i_mut,)
            if i < len(a):
                match_value: __A | None = f(a[i])
                if match_value is None:
                    i_mut = i + 1
                    continue

                else: 
                    return some(value_1(match_value))


            else: 
                return None

            break

    return loop(0)


def ResizeArray_zip(a: Array[Any], b: Array[Any]) -> Array[tuple[_T, _U]]:
    c: Array[tuple[_T, _U]] = []
    def _arrow645(x: int, y: int, a: Any=a, b: Any=b) -> int:
        return compare_primitives(x, y)

    n: int = min(_arrow645, len(a), len(b)) or 0
    for i in range(0, (n - 1) + 1, 1):
        (c.append((a[i], b[i])))
    return c


def ResizeArray_tryFind(f: Callable[[_T], bool], a: Array[Any]) -> Any | None:
    def loop(i_mut: int, f: Any=f, a: Any=a) -> _T | None:
        while True:
            (i,) = (i_mut,)
            if i < len(a):
                if f(a[i]):
                    return some(a[i])

                else: 
                    i_mut = i + 1
                    continue


            else: 
                return None

            break

    return loop(0)


__all__ = ["Option_fromValueWithDefault", "Option_mapDefault", "Option_mapOrDefault", "Option_fromSeq", "List_tryPickAndRemove", "Dictionary_addOrUpdate", "Dictionary_ofSeq", "Dictionary_tryFind", "Dictionary_ofSeqWithMerge", "Dictionary_init", "Dictionary_items", "StringDictionary_ofSeq", "IntDictionary_ofSeq", "ResizeArray_create", "ResizeArray_singleton", "ResizeArray_map", "ResizeArray_mapi", "ResizeArray_map2", "ResizeArray_choose", "ResizeArray_filter", "ResizeArray_fold", "ResizeArray_foldBack", "ResizeArray_iter", "ResizeArray_iteri", "ResizeArray_init", "ResizeArray_reduce", "ResizeArray_collect", "ResizeArray_distinct", "ResizeArray_isEmpty", "ResizeArray_append", "ResizeArray_appendSingleton", "ResizeArray_indexed", "ResizeArray_rev", "ResizeArray_take", "ResizeArray_groupBy", "ResizeArray_tryPick", "ResizeArray_zip", "ResizeArray_tryFind"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.ontology_source_reference import OntologySourceReference
from ...arctrl_json.encode import default_spaces
from ...arctrl_json.ontology_source_reference import (decoder as decoder_1, encoder, ROCrate_decoder, ROCrate_encoder, ISAJson_decoder, ISAJson_encoder)
from ...fable_library.result import FSharpResult_2
from ...fable_library.string_ import (to_text, printf)
from ...thoth_json_core.types import IEncodable
from ...thoth_json_python.decode import Decode_fromString
from ...thoth_json_python.encode import to_string

def ARCtrl_OntologySourceReference__OntologySourceReference_fromJsonString_Static_Z721C83C5(s: str) -> OntologySourceReference:
    match_value: FSharpResult_2[OntologySourceReference, str] = Decode_fromString(decoder_1, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_OntologySourceReference__OntologySourceReference_toJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[OntologySourceReference], str]:
    def _arrow3789(obj: OntologySourceReference, spaces: Any=spaces) -> str:
        value: IEncodable = encoder(obj)
        return to_string(default_spaces(spaces), value)

    return _arrow3789


def ARCtrl_OntologySourceReference__OntologySourceReference_ToJsonString_71136F3F(this: OntologySourceReference, spaces: int | None=None) -> str:
    return ARCtrl_OntologySourceReference__OntologySourceReference_toJsonString_Static_71136F3F(spaces)(this)


def ARCtrl_OntologySourceReference__OntologySourceReference_fromROCrateJsonString_Static_Z721C83C5(s: str) -> OntologySourceReference:
    match_value: FSharpResult_2[OntologySourceReference, str] = Decode_fromString(ROCrate_decoder, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_OntologySourceReference__OntologySourceReference_toROCrateJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[OntologySourceReference], str]:
    def _arrow3790(obj: OntologySourceReference, spaces: Any=spaces) -> str:
        value: IEncodable = ROCrate_encoder(obj)
        return to_string(default_spaces(spaces), value)

    return _arrow3790


def ARCtrl_OntologySourceReference__OntologySourceReference_ToROCrateJsonString_71136F3F(this: OntologySourceReference, spaces: int | None=None) -> str:
    return ARCtrl_OntologySourceReference__OntologySourceReference_toROCrateJsonString_Static_71136F3F(spaces)(this)


def ARCtrl_OntologySourceReference__OntologySourceReference_fromISAJsonString_Static_Z721C83C5(s: str) -> OntologySourceReference:
    match_value: FSharpResult_2[OntologySourceReference, str] = Decode_fromString(ISAJson_decoder, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_OntologySourceReference__OntologySourceReference_toISAJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[OntologySourceReference], str]:
    def _arrow3791(obj: OntologySourceReference, spaces: Any=spaces) -> str:
        value: IEncodable = ISAJson_encoder(None, obj)
        return to_string(default_spaces(spaces), value)

    return _arrow3791


def ARCtrl_OntologySourceReference__OntologySourceReference_ToISAJsonString_71136F3F(this: OntologySourceReference, spaces: int | None=None) -> str:
    return ARCtrl_OntologySourceReference__OntologySourceReference_toISAJsonString_Static_71136F3F(spaces)(this)


__all__ = ["ARCtrl_OntologySourceReference__OntologySourceReference_fromJsonString_Static_Z721C83C5", "ARCtrl_OntologySourceReference__OntologySourceReference_toJsonString_Static_71136F3F", "ARCtrl_OntologySourceReference__OntologySourceReference_ToJsonString_71136F3F", "ARCtrl_OntologySourceReference__OntologySourceReference_fromROCrateJsonString_Static_Z721C83C5", "ARCtrl_OntologySourceReference__OntologySourceReference_toROCrateJsonString_Static_71136F3F", "ARCtrl_OntologySourceReference__OntologySourceReference_ToROCrateJsonString_71136F3F", "ARCtrl_OntologySourceReference__OntologySourceReference_fromISAJsonString_Static_Z721C83C5", "ARCtrl_OntologySourceReference__OntologySourceReference_toISAJsonString_Static_71136F3F", "ARCtrl_OntologySourceReference__OntologySourceReference_ToISAJsonString_71136F3F"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ..arctrl_core.arc_types import (ArcAssay, ArcStudy)
from ..arctrl_core.comment import Comment
from ..arctrl_core.conversion import (ARCtrl_ArcTables__ArcTables_GetProcesses, ARCtrl_ArcTables__ArcTables_fromProcesses_Static_62A3309D, Person_setSourceAssayComment, Person_getSourceAssayIdentifiersFromComments, Person_removeSourceAssayComments)
from ..arctrl_core.data import Data
from ..arctrl_core.data_map import DataMap
from ..arctrl_core.Helper.collections_ import (ResizeArray_map, Option_fromValueWithDefault)
from ..arctrl_core.Helper.identifier import (Study_tryFileNameFromIdentifier, Study_tryIdentifierFromFileName, create_missing_identifier, Study_fileNameFromIdentifier)
from ..arctrl_core.ontology_annotation import OntologyAnnotation
from ..arctrl_core.person import Person
from ..arctrl_core.Process.factor import Factor
from ..arctrl_core.Process.material_attribute import MaterialAttribute
from ..arctrl_core.Process.process import Process
from ..arctrl_core.Process.process_sequence import (get_data, get_units, get_factors, get_characteristics, get_protocols)
from ..arctrl_core.Process.protocol import Protocol
from ..arctrl_core.publication import Publication
from ..arctrl_core.Table.arc_table import ArcTable
from ..arctrl_core.Table.arc_tables import ArcTables
from ..arctrl_core.Table.composite_cell import CompositeCell
from ..fable_library.list import (try_find, FSharpList, choose, of_array, singleton, map as map_2, empty)
from ..fable_library.option import (default_arg, value as value_17, map, bind, default_arg_with)
from ..fable_library.seq import (map as map_1, is_empty)
from ..fable_library.string_ import replace
from ..fable_library.types import Array
from ..fable_library.util import (IEnumerable_1, get_enumerator, dispose)
from ..thoth_json_core.decode import (object, IRequiredGetter, string, IOptionalGetter, resize_array, IGetters, list_1 as list_1_2)
from ..thoth_json_core.encode import list_1 as list_1_1
from ..thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from .assay import (ROCrate_encoder as ROCrate_encoder_4, ROCrate_decoder as ROCrate_decoder_1, ISAJson_encoder as ISAJson_encoder_5, ISAJson_decoder as ISAJson_decoder_2)
from .comment import (encoder as encoder_9, decoder as decoder_6, ROCrate_encoder as ROCrate_encoder_6, ROCrate_decoder as ROCrate_decoder_5, ISAJson_encoder as ISAJson_encoder_6, ISAJson_decoder as ISAJson_decoder_5)
from .context.rocrate.isa_study_context import context_jsonvalue
from .data import ROCrate_encoder as ROCrate_encoder_5
from .DataMap.data_map import (encoder as encoder_8, decoder as decoder_5, encoder_compressed as encoder_compressed_2, decoder_compressed as decoder_compressed_2)
from .decode import Decode_objectNoAdditionalProperties
from .encode import (try_include, try_include_seq, try_include_list)
from .idtable import encode
from .ontology_annotation import (OntologyAnnotation_encoder, OntologyAnnotation_decoder, OntologyAnnotation_ROCrate_encoderDefinedTerm, OntologyAnnotation_ROCrate_decoderDefinedTerm, OntologyAnnotation_ISAJson_encoder, OntologyAnnotation_ISAJson_decoder)
from .person import (encoder as encoder_6, decoder as decoder_3, ROCrate_encoder as ROCrate_encoder_2, ROCrate_decoder as ROCrate_decoder_4, ISAJson_encoder as ISAJson_encoder_3, ISAJson_decoder as ISAJson_decoder_3)
from .Process.factor import encoder as encoder_10
from .Process.material_attribute import encoder as encoder_11
from .Process.process import (ROCrate_encoder as ROCrate_encoder_3, ROCrate_decoder as ROCrate_decoder_2, ISAJson_encoder as ISAJson_encoder_4, ISAJson_decoder as ISAJson_decoder_1)
from .Process.protocol import ISAJson_encoder as ISAJson_encoder_1
from .Process.study_materials import encoder as encoder_12
from .publication import (encoder as encoder_5, decoder as decoder_2, ROCrate_encoder as ROCrate_encoder_1, ROCrate_decoder as ROCrate_decoder_3, ISAJson_encoder as ISAJson_encoder_2, ISAJson_decoder as ISAJson_decoder_4)
from .Table.arc_table import (encoder as encoder_7, decoder as decoder_4, encoder_compressed as encoder_compressed_1, decoder_compressed as decoder_compressed_1)

__A_ = TypeVar("__A_")

def Helper_getAssayInformation(assays: FSharpList[ArcAssay] | None, study: ArcStudy) -> Array[ArcAssay]:
    if assays is not None:
        def f(assay_id: str, assays: Any=assays, study: Any=study) -> ArcAssay:
            def predicate(a: ArcAssay, assay_id: Any=assay_id) -> bool:
                return a.Identifier == assay_id

            return default_arg(try_find(predicate, value_17(assays)), ArcAssay.init(assay_id))

        return ResizeArray_map(f, study.RegisteredAssayIdentifiers)

    else: 
        return study.GetRegisteredAssaysOrIdentifier()



def encoder(study: ArcStudy) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], study: Any=study) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow3062(__unit: None=None, study: Any=study) -> IEncodable:
        value: str = study.Identifier
        class ObjectExpr3061(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr3061()

    def _arrow3064(value_1: str, study: Any=study) -> IEncodable:
        class ObjectExpr3063(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_1)

        return ObjectExpr3063()

    def _arrow3066(value_3: str, study: Any=study) -> IEncodable:
        class ObjectExpr3065(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_3)

        return ObjectExpr3065()

    def _arrow3068(value_5: str, study: Any=study) -> IEncodable:
        class ObjectExpr3067(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_5)

        return ObjectExpr3067()

    def _arrow3070(value_7: str, study: Any=study) -> IEncodable:
        class ObjectExpr3069(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_7)

        return ObjectExpr3069()

    def _arrow3071(oa: Publication, study: Any=study) -> IEncodable:
        return encoder_5(oa)

    def _arrow3072(person: Person, study: Any=study) -> IEncodable:
        return encoder_6(person)

    def _arrow3073(oa_1: OntologyAnnotation, study: Any=study) -> IEncodable:
        return OntologyAnnotation_encoder(oa_1)

    def _arrow3074(table: ArcTable, study: Any=study) -> IEncodable:
        return encoder_7(table)

    def _arrow3075(dm: DataMap, study: Any=study) -> IEncodable:
        return encoder_8(dm)

    def _arrow3077(value_9: str, study: Any=study) -> IEncodable:
        class ObjectExpr3076(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                return helpers_5.encode_string(value_9)

        return ObjectExpr3076()

    def _arrow3078(comment: Comment, study: Any=study) -> IEncodable:
        return encoder_9(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("Identifier", _arrow3062()), try_include("Title", _arrow3064, study.Title), try_include("Description", _arrow3066, study.Description), try_include("SubmissionDate", _arrow3068, study.SubmissionDate), try_include("PublicReleaseDate", _arrow3070, study.PublicReleaseDate), try_include_seq("Publications", _arrow3071, study.Publications), try_include_seq("Contacts", _arrow3072, study.Contacts), try_include_seq("StudyDesignDescriptors", _arrow3073, study.StudyDesignDescriptors), try_include_seq("Tables", _arrow3074, study.Tables), try_include("DataMap", _arrow3075, study.DataMap), try_include_seq("RegisteredAssayIdentifiers", _arrow3077, study.RegisteredAssayIdentifiers), try_include_seq("Comments", _arrow3078, study.Comments)]))
    class ObjectExpr3079(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any], study: Any=study) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_6))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_6.encode_object(arg)

    return ObjectExpr3079()


def _arrow3094(get: IGetters) -> ArcStudy:
    def _arrow3080(__unit: None=None) -> str:
        object_arg: IRequiredGetter = get.Required
        return object_arg.Field("Identifier", string)

    def _arrow3081(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("Title", string)

    def _arrow3082(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("Description", string)

    def _arrow3083(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("SubmissionDate", string)

    def _arrow3084(__unit: None=None) -> str | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("PublicReleaseDate", string)

    def _arrow3085(__unit: None=None) -> Array[Publication] | None:
        arg_11: Decoder_1[Array[Publication]] = resize_array(decoder_2)
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("Publications", arg_11)

    def _arrow3086(__unit: None=None) -> Array[Person] | None:
        arg_13: Decoder_1[Array[Person]] = resize_array(decoder_3)
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("Contacts", arg_13)

    def _arrow3087(__unit: None=None) -> Array[OntologyAnnotation] | None:
        arg_15: Decoder_1[Array[OntologyAnnotation]] = resize_array(OntologyAnnotation_decoder)
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("StudyDesignDescriptors", arg_15)

    def _arrow3088(__unit: None=None) -> Array[ArcTable] | None:
        arg_17: Decoder_1[Array[ArcTable]] = resize_array(decoder_4)
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("Tables", arg_17)

    def _arrow3089(__unit: None=None) -> DataMap | None:
        object_arg_9: IOptionalGetter = get.Optional
        return object_arg_9.Field("DataMap", decoder_5)

    def _arrow3092(__unit: None=None) -> Array[str] | None:
        arg_21: Decoder_1[Array[str]] = resize_array(string)
        object_arg_10: IOptionalGetter = get.Optional
        return object_arg_10.Field("RegisteredAssayIdentifiers", arg_21)

    def _arrow3093(__unit: None=None) -> Array[Comment] | None:
        arg_23: Decoder_1[Array[Comment]] = resize_array(decoder_6)
        object_arg_11: IOptionalGetter = get.Optional
        return object_arg_11.Field("Comments", arg_23)

    return ArcStudy(_arrow3080(), _arrow3081(), _arrow3082(), _arrow3083(), _arrow3084(), _arrow3085(), _arrow3086(), _arrow3087(), _arrow3088(), _arrow3089(), _arrow3092(), _arrow3093())


decoder: Decoder_1[ArcStudy] = object(_arrow3094)

def encoder_compressed(string_table: Any, oa_table: Any, cell_table: Any, study: ArcStudy) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, study: Any=study) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow3107(__unit: None=None, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, study: Any=study) -> IEncodable:
        value: str = study.Identifier
        class ObjectExpr3106(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr3106()

    def _arrow3110(value_1: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, study: Any=study) -> IEncodable:
        class ObjectExpr3109(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_1)

        return ObjectExpr3109()

    def _arrow3117(value_3: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, study: Any=study) -> IEncodable:
        class ObjectExpr3116(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_3)

        return ObjectExpr3116()

    def _arrow3119(value_5: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, study: Any=study) -> IEncodable:
        class ObjectExpr3118(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_5)

        return ObjectExpr3118()

    def _arrow3122(value_7: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, study: Any=study) -> IEncodable:
        class ObjectExpr3121(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_7)

        return ObjectExpr3121()

    def _arrow3123(oa: Publication, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, study: Any=study) -> IEncodable:
        return encoder_5(oa)

    def _arrow3124(person: Person, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, study: Any=study) -> IEncodable:
        return encoder_6(person)

    def _arrow3125(oa_1: OntologyAnnotation, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, study: Any=study) -> IEncodable:
        return OntologyAnnotation_encoder(oa_1)

    def _arrow3126(table: ArcTable, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, study: Any=study) -> IEncodable:
        return encoder_compressed_1(string_table, oa_table, cell_table, table)

    def _arrow3127(dm: DataMap, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, study: Any=study) -> IEncodable:
        return encoder_compressed_2(string_table, oa_table, cell_table, dm)

    def _arrow3129(value_9: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, study: Any=study) -> IEncodable:
        class ObjectExpr3128(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                return helpers_5.encode_string(value_9)

        return ObjectExpr3128()

    def _arrow3131(comment: Comment, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, study: Any=study) -> IEncodable:
        return encoder_9(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("Identifier", _arrow3107()), try_include("Title", _arrow3110, study.Title), try_include("Description", _arrow3117, study.Description), try_include("SubmissionDate", _arrow3119, study.SubmissionDate), try_include("PublicReleaseDate", _arrow3122, study.PublicReleaseDate), try_include_seq("Publications", _arrow3123, study.Publications), try_include_seq("Contacts", _arrow3124, study.Contacts), try_include_seq("StudyDesignDescriptors", _arrow3125, study.StudyDesignDescriptors), try_include_seq("Tables", _arrow3126, study.Tables), try_include("DataMap", _arrow3127, study.DataMap), try_include_seq("RegisteredAssayIdentifiers", _arrow3129, study.RegisteredAssayIdentifiers), try_include_seq("Comments", _arrow3131, study.Comments)]))
    class ObjectExpr3138(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any], string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, study: Any=study) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_6))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_6.encode_object(arg)

    return ObjectExpr3138()


def decoder_compressed(string_table: Array[str], oa_table: Array[OntologyAnnotation], cell_table: Array[CompositeCell]) -> Decoder_1[ArcStudy]:
    def _arrow3173(get: IGetters, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table) -> ArcStudy:
        def _arrow3144(__unit: None=None) -> str:
            object_arg: IRequiredGetter = get.Required
            return object_arg.Field("Identifier", string)

        def _arrow3145(__unit: None=None) -> str | None:
            object_arg_1: IOptionalGetter = get.Optional
            return object_arg_1.Field("Title", string)

        def _arrow3147(__unit: None=None) -> str | None:
            object_arg_2: IOptionalGetter = get.Optional
            return object_arg_2.Field("Description", string)

        def _arrow3149(__unit: None=None) -> str | None:
            object_arg_3: IOptionalGetter = get.Optional
            return object_arg_3.Field("SubmissionDate", string)

        def _arrow3152(__unit: None=None) -> str | None:
            object_arg_4: IOptionalGetter = get.Optional
            return object_arg_4.Field("PublicReleaseDate", string)

        def _arrow3154(__unit: None=None) -> Array[Publication] | None:
            arg_11: Decoder_1[Array[Publication]] = resize_array(decoder_2)
            object_arg_5: IOptionalGetter = get.Optional
            return object_arg_5.Field("Publications", arg_11)

        def _arrow3156(__unit: None=None) -> Array[Person] | None:
            arg_13: Decoder_1[Array[Person]] = resize_array(decoder_3)
            object_arg_6: IOptionalGetter = get.Optional
            return object_arg_6.Field("Contacts", arg_13)

        def _arrow3159(__unit: None=None) -> Array[OntologyAnnotation] | None:
            arg_15: Decoder_1[Array[OntologyAnnotation]] = resize_array(OntologyAnnotation_decoder)
            object_arg_7: IOptionalGetter = get.Optional
            return object_arg_7.Field("StudyDesignDescriptors", arg_15)

        def _arrow3164(__unit: None=None) -> Array[ArcTable] | None:
            arg_17: Decoder_1[Array[ArcTable]] = resize_array(decoder_compressed_1(string_table, oa_table, cell_table))
            object_arg_8: IOptionalGetter = get.Optional
            return object_arg_8.Field("Tables", arg_17)

        def _arrow3165(__unit: None=None) -> DataMap | None:
            arg_19: Decoder_1[DataMap] = decoder_compressed_2(string_table, oa_table, cell_table)
            object_arg_9: IOptionalGetter = get.Optional
            return object_arg_9.Field("DataMap", arg_19)

        def _arrow3168(__unit: None=None) -> Array[str] | None:
            arg_21: Decoder_1[Array[str]] = resize_array(string)
            object_arg_10: IOptionalGetter = get.Optional
            return object_arg_10.Field("RegisteredAssayIdentifiers", arg_21)

        def _arrow3171(__unit: None=None) -> Array[Comment] | None:
            arg_23: Decoder_1[Array[Comment]] = resize_array(decoder_6)
            object_arg_11: IOptionalGetter = get.Optional
            return object_arg_11.Field("Comments", arg_23)

        return ArcStudy(_arrow3144(), _arrow3145(), _arrow3147(), _arrow3149(), _arrow3152(), _arrow3154(), _arrow3156(), _arrow3159(), _arrow3164(), _arrow3165(), _arrow3168(), _arrow3171())

    return object(_arrow3173)


def ROCrate_genID(a: ArcStudy) -> str:
    match_value: str = a.Identifier
    if match_value == "":
        return "#EmptyStudy"

    else: 
        return ("studies/" + replace(match_value, " ", "_")) + "/"



def ROCrate_encoder(assays: FSharpList[ArcAssay] | None, s: ArcStudy) -> IEncodable:
    file_name: str | None = Study_tryFileNameFromIdentifier(s.Identifier)
    processes: FSharpList[Process] = ARCtrl_ArcTables__ArcTables_GetProcesses(s)
    assays_1: Array[ArcAssay] = Helper_getAssayInformation(assays, s)
    def chooser(tupled_arg: tuple[str, IEncodable | None], assays: Any=assays, s: Any=s) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow3186(__unit: None=None, assays: Any=assays, s: Any=s) -> IEncodable:
        value: str = ROCrate_genID(s)
        class ObjectExpr3185(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr3185()

    class ObjectExpr3189(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], assays: Any=assays, s: Any=s) -> Any:
            return helpers_1.encode_string("Study")

    class ObjectExpr3192(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any], assays: Any=assays, s: Any=s) -> Any:
            return helpers_2.encode_string("Study")

    def _arrow3194(__unit: None=None, assays: Any=assays, s: Any=s) -> IEncodable:
        value_3: str = s.Identifier
        class ObjectExpr3193(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_3)

        return ObjectExpr3193()

    def _arrow3196(value_4: str, assays: Any=assays, s: Any=s) -> IEncodable:
        class ObjectExpr3195(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_4)

        return ObjectExpr3195()

    def _arrow3198(value_6: str, assays: Any=assays, s: Any=s) -> IEncodable:
        class ObjectExpr3197(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                return helpers_5.encode_string(value_6)

        return ObjectExpr3197()

    def _arrow3200(value_8: str, assays: Any=assays, s: Any=s) -> IEncodable:
        class ObjectExpr3199(IEncodable):
            def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
                return helpers_6.encode_string(value_8)

        return ObjectExpr3199()

    def _arrow3201(oa: OntologyAnnotation, assays: Any=assays, s: Any=s) -> IEncodable:
        return OntologyAnnotation_ROCrate_encoderDefinedTerm(oa)

    def _arrow3203(value_10: str, assays: Any=assays, s: Any=s) -> IEncodable:
        class ObjectExpr3202(IEncodable):
            def Encode(self, helpers_7: IEncoderHelpers_1[Any]) -> Any:
                return helpers_7.encode_string(value_10)

        return ObjectExpr3202()

    def _arrow3205(value_12: str, assays: Any=assays, s: Any=s) -> IEncodable:
        class ObjectExpr3204(IEncodable):
            def Encode(self, helpers_8: IEncoderHelpers_1[Any]) -> Any:
                return helpers_8.encode_string(value_12)

        return ObjectExpr3204()

    def _arrow3206(oa_1: Publication, assays: Any=assays, s: Any=s) -> IEncodable:
        return ROCrate_encoder_1(oa_1)

    def _arrow3207(oa_2: Person, assays: Any=assays, s: Any=s) -> IEncodable:
        return ROCrate_encoder_2(oa_2)

    def _arrow3209(__unit: None=None, assays: Any=assays, s: Any=s) -> Callable[[Process], IEncodable]:
        study_name: str | None = s.Identifier
        def _arrow3208(oa_3: Process) -> IEncodable:
            return ROCrate_encoder_3(study_name, None, oa_3)

        return _arrow3208

    def _arrow3211(__unit: None=None, assays: Any=assays, s: Any=s) -> Callable[[ArcAssay], IEncodable]:
        assay_name_1: str | None = s.Identifier
        def _arrow3210(a_1: ArcAssay) -> IEncodable:
            return ROCrate_encoder_4(assay_name_1, a_1)

        return _arrow3210

    def _arrow3212(oa_4: Data, assays: Any=assays, s: Any=s) -> IEncodable:
        return ROCrate_encoder_5(oa_4)

    def _arrow3213(comment: Comment, assays: Any=assays, s: Any=s) -> IEncodable:
        return ROCrate_encoder_6(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow3186()), ("@type", list_1_1(singleton(ObjectExpr3189()))), ("additionalType", ObjectExpr3192()), ("identifier", _arrow3194()), try_include("filename", _arrow3196, file_name), try_include("title", _arrow3198, s.Title), try_include("description", _arrow3200, s.Description), try_include_seq("studyDesignDescriptors", _arrow3201, s.StudyDesignDescriptors), try_include("submissionDate", _arrow3203, s.SubmissionDate), try_include("publicReleaseDate", _arrow3205, s.PublicReleaseDate), try_include_seq("publications", _arrow3206, s.Publications), try_include_seq("people", _arrow3207, s.Contacts), try_include_list("processSequence", _arrow3209(), processes), try_include_seq("assays", _arrow3211(), assays_1), try_include_list("dataFiles", _arrow3212, get_data(processes)), try_include_seq("comments", _arrow3213, s.Comments), ("@context", context_jsonvalue)]))
    class ObjectExpr3214(IEncodable):
        def Encode(self, helpers_9: IEncoderHelpers_1[Any], assays: Any=assays, s: Any=s) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_9))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_9.encode_object(arg)

    return ObjectExpr3214()


def _arrow3225(get: IGetters) -> tuple[ArcStudy, FSharpList[ArcAssay]]:
    def _arrow3215(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("filename", string)

    identifier: str = default_arg(bind(Study_tryIdentifierFromFileName, _arrow3215()), create_missing_identifier())
    assays: FSharpList[ArcAssay] | None
    arg_3: Decoder_1[FSharpList[ArcAssay]] = list_1_2(ROCrate_decoder_1)
    object_arg_1: IOptionalGetter = get.Optional
    assays = object_arg_1.Field("assays", arg_3)
    def mapping_1(arg_4: FSharpList[ArcAssay]) -> Array[str]:
        def mapping(a: ArcAssay, arg_4: Any=arg_4) -> str:
            return a.Identifier

        return list(map_2(mapping, arg_4))

    assay_identifiers: Array[str] | None = map(mapping_1, assays)
    def mapping_2(ps: FSharpList[Process]) -> Array[ArcTable]:
        return ARCtrl_ArcTables__ArcTables_fromProcesses_Static_62A3309D(ps).Tables

    def _arrow3216(__unit: None=None) -> FSharpList[Process] | None:
        arg_6: Decoder_1[FSharpList[Process]] = list_1_2(ROCrate_decoder_2)
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("processSequence", arg_6)

    tables: Array[ArcTable] | None = map(mapping_2, _arrow3216())
    def _arrow3217(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("title", string)

    def _arrow3218(__unit: None=None) -> str | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("description", string)

    def _arrow3219(__unit: None=None) -> str | None:
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("submissionDate", string)

    def _arrow3220(__unit: None=None) -> str | None:
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("publicReleaseDate", string)

    def _arrow3221(__unit: None=None) -> Array[Publication] | None:
        arg_16: Decoder_1[Array[Publication]] = resize_array(ROCrate_decoder_3)
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("publications", arg_16)

    def _arrow3222(__unit: None=None) -> Array[Person] | None:
        arg_18: Decoder_1[Array[Person]] = resize_array(ROCrate_decoder_4)
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("people", arg_18)

    def _arrow3223(__unit: None=None) -> Array[OntologyAnnotation] | None:
        arg_20: Decoder_1[Array[OntologyAnnotation]] = resize_array(OntologyAnnotation_ROCrate_decoderDefinedTerm)
        object_arg_9: IOptionalGetter = get.Optional
        return object_arg_9.Field("studyDesignDescriptors", arg_20)

    def _arrow3224(__unit: None=None) -> Array[Comment] | None:
        arg_22: Decoder_1[Array[Comment]] = resize_array(ROCrate_decoder_5)
        object_arg_10: IOptionalGetter = get.Optional
        return object_arg_10.Field("comments", arg_22)

    return (ArcStudy(identifier, _arrow3217(), _arrow3218(), _arrow3219(), _arrow3220(), _arrow3221(), _arrow3222(), _arrow3223(), tables, None, assay_identifiers, _arrow3224()), default_arg(assays, empty()))


ROCrate_decoder: Decoder_1[tuple[ArcStudy, FSharpList[ArcAssay]]] = object(_arrow3225)

def ISAJson_encoder(id_map: Any | None, assays: FSharpList[ArcAssay] | None, s: ArcStudy) -> IEncodable:
    def f(s_1: ArcStudy, id_map: Any=id_map, assays: Any=assays, s: Any=s) -> IEncodable:
        study: ArcStudy = s_1.Copy(True)
        file_name: str = Study_fileNameFromIdentifier(study.Identifier)
        assays_1: Array[ArcAssay]
        n: Array[ArcAssay] = []
        enumerator: Any = get_enumerator(Helper_getAssayInformation(assays, study))
        try: 
            while enumerator.System_Collections_IEnumerator_MoveNext():
                a: ArcAssay = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                assay: ArcAssay = a.Copy()
                enumerator_1: Any = get_enumerator(assay.Performers)
                try: 
                    while enumerator_1.System_Collections_IEnumerator_MoveNext():
                        person_1: Person = Person_setSourceAssayComment(enumerator_1.System_Collections_Generic_IEnumerator_1_get_Current(), assay.Identifier)
                        (study.Contacts.append(person_1))

                finally: 
                    dispose(enumerator_1)

                assay.Performers = []
                (n.append(assay))

        finally: 
            dispose(enumerator)

        assays_1 = n
        processes: FSharpList[Process] = ARCtrl_ArcTables__ArcTables_GetProcesses(study)
        def encoder_1(oa: OntologyAnnotation, s_1: Any=s_1) -> IEncodable:
            return OntologyAnnotation_ISAJson_encoder(id_map, oa)

        encoded_units: tuple[str, IEncodable | None] = try_include_list("unitCategories", encoder_1, get_units(processes))
        def encoder_2(value_1: Factor, s_1: Any=s_1) -> IEncodable:
            return encoder_10(id_map, value_1)

        encoded_factors: tuple[str, IEncodable | None] = try_include_list("factors", encoder_2, get_factors(processes))
        def encoder_3(value_3: MaterialAttribute, s_1: Any=s_1) -> IEncodable:
            return encoder_11(id_map, value_3)

        encoded_characteristics: tuple[str, IEncodable | None] = try_include_list("characteristicCategories", encoder_3, get_characteristics(processes))
        def _arrow3226(ps: FSharpList[Process], s_1: Any=s_1) -> IEncodable:
            return encoder_12(id_map, ps)

        encoded_materials: tuple[str, IEncodable | None] = try_include("materials", _arrow3226, Option_fromValueWithDefault(empty(), processes))
        encoded_protocols: tuple[str, IEncodable | None]
        value_5: FSharpList[Protocol] = get_protocols(processes)
        def _arrow3228(__unit: None=None, s_1: Any=s_1) -> Callable[[Protocol], IEncodable]:
            study_name: str | None = s_1.Identifier
            def _arrow3227(oa_1: Protocol) -> IEncodable:
                return ISAJson_encoder_1(study_name, None, None, id_map, oa_1)

            return _arrow3227

        encoded_protocols = try_include_list("protocols", _arrow3228(), value_5)
        def chooser(tupled_arg: tuple[str, IEncodable | None], s_1: Any=s_1) -> tuple[str, IEncodable] | None:
            def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
                return (tupled_arg[0], v_1)

            return map(mapping, tupled_arg[1])

        def _arrow3232(__unit: None=None, s_1: Any=s_1) -> IEncodable:
            value_6: str = ROCrate_genID(study)
            class ObjectExpr3231(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value_6)

            return ObjectExpr3231()

        class ObjectExpr3233(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any], s_1: Any=s_1) -> Any:
                return helpers_1.encode_string(file_name)

        def _arrow3235(__unit: None=None, s_1: Any=s_1) -> IEncodable:
            value_8: str = study.Identifier
            class ObjectExpr3234(IEncodable):
                def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_2.encode_string(value_8)

            return ObjectExpr3234()

        def _arrow3237(value_9: str, s_1: Any=s_1) -> IEncodable:
            class ObjectExpr3236(IEncodable):
                def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_3.encode_string(value_9)

            return ObjectExpr3236()

        def _arrow3239(value_11: str, s_1: Any=s_1) -> IEncodable:
            class ObjectExpr3238(IEncodable):
                def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_4.encode_string(value_11)

            return ObjectExpr3238()

        def _arrow3241(value_13: str, s_1: Any=s_1) -> IEncodable:
            class ObjectExpr3240(IEncodable):
                def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_5.encode_string(value_13)

            return ObjectExpr3240()

        def _arrow3243(value_15: str, s_1: Any=s_1) -> IEncodable:
            class ObjectExpr3242(IEncodable):
                def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_6.encode_string(value_15)

            return ObjectExpr3242()

        def _arrow3244(oa_2: Publication, s_1: Any=s_1) -> IEncodable:
            return ISAJson_encoder_2(id_map, oa_2)

        def _arrow3245(person_2: Person, s_1: Any=s_1) -> IEncodable:
            return ISAJson_encoder_3(id_map, person_2)

        def _arrow3246(oa_3: OntologyAnnotation, s_1: Any=s_1) -> IEncodable:
            return OntologyAnnotation_ISAJson_encoder(id_map, oa_3)

        def _arrow3248(__unit: None=None, s_1: Any=s_1) -> Callable[[Process], IEncodable]:
            study_name_1: str | None = s_1.Identifier
            def _arrow3247(oa_4: Process) -> IEncodable:
                return ISAJson_encoder_4(study_name_1, None, id_map, oa_4)

            return _arrow3247

        def _arrow3250(__unit: None=None, s_1: Any=s_1) -> Callable[[ArcAssay], IEncodable]:
            study_name_2: str | None = s_1.Identifier
            def _arrow3249(a_2: ArcAssay) -> IEncodable:
                return ISAJson_encoder_5(study_name_2, id_map, a_2)

            return _arrow3249

        def _arrow3251(comment: Comment, s_1: Any=s_1) -> IEncodable:
            return ISAJson_encoder_6(id_map, comment)

        values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow3232()), ("filename", ObjectExpr3233()), ("identifier", _arrow3235()), try_include("title", _arrow3237, study.Title), try_include("description", _arrow3239, study.Description), try_include("submissionDate", _arrow3241, study.SubmissionDate), try_include("publicReleaseDate", _arrow3243, study.PublicReleaseDate), try_include_seq("publications", _arrow3244, study.Publications), try_include_seq("people", _arrow3245, study.Contacts), try_include_seq("studyDesignDescriptors", _arrow3246, study.StudyDesignDescriptors), encoded_protocols, encoded_materials, try_include_list("processSequence", _arrow3248(), processes), try_include_seq("assays", _arrow3250(), assays_1), encoded_factors, encoded_characteristics, encoded_units, try_include_seq("comments", _arrow3251, study.Comments)]))
        class ObjectExpr3252(IEncodable):
            def Encode(self, helpers_7: IEncoderHelpers_1[Any], s_1: Any=s_1) -> Any:
                def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_7))

                arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
                return helpers_7.encode_object(arg)

        return ObjectExpr3252()

    if id_map is not None:
        def _arrow3253(s_2: ArcStudy, id_map: Any=id_map, assays: Any=assays, s: Any=s) -> str:
            return ROCrate_genID(s_2)

        return encode(_arrow3253, f, s, id_map)

    else: 
        return f(s)



ISAJson_allowedFields: FSharpList[str] = of_array(["@id", "filename", "identifier", "title", "description", "submissionDate", "publicReleaseDate", "publications", "people", "studyDesignDescriptors", "protocols", "materials", "assays", "factors", "characteristicCategories", "unitCategories", "processSequence", "comments", "@type", "@context"])

def _arrow3264(get: IGetters) -> tuple[ArcStudy, FSharpList[ArcAssay]]:
    def _arrow3254(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("identifier", string)

    def def_thunk(__unit: None=None) -> str:
        def _arrow3255(__unit: None=None) -> str | None:
            object_arg_1: IOptionalGetter = get.Optional
            return object_arg_1.Field("filename", string)

        return default_arg(bind(Study_tryIdentifierFromFileName, _arrow3255()), create_missing_identifier())

    identifier: str = default_arg_with(_arrow3254(), def_thunk)
    def mapping(arg_6: FSharpList[Process]) -> Array[ArcTable]:
        a: ArcTables = ARCtrl_ArcTables__ArcTables_fromProcesses_Static_62A3309D(arg_6)
        return a.Tables

    def _arrow3256(__unit: None=None) -> FSharpList[Process] | None:
        arg_5: Decoder_1[FSharpList[Process]] = list_1_2(ISAJson_decoder_1)
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("processSequence", arg_5)

    tables: Array[ArcTable] | None = map(mapping, _arrow3256())
    assays: FSharpList[ArcAssay] | None
    arg_8: Decoder_1[FSharpList[ArcAssay]] = list_1_2(ISAJson_decoder_2)
    object_arg_3: IOptionalGetter = get.Optional
    assays = object_arg_3.Field("assays", arg_8)
    persons_raw: Array[Person] | None
    arg_10: Decoder_1[Array[Person]] = resize_array(ISAJson_decoder_3)
    object_arg_4: IOptionalGetter = get.Optional
    persons_raw = object_arg_4.Field("people", arg_10)
    persons: Array[Person] = []
    if persons_raw is not None:
        enumerator: Any = get_enumerator(value_17(persons_raw))
        try: 
            while enumerator.System_Collections_IEnumerator_MoveNext():
                person: Person = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                source_assays: IEnumerable_1[str] = Person_getSourceAssayIdentifiersFromComments(person)
                with get_enumerator(source_assays) as enumerator_1:
                    while enumerator_1.System_Collections_IEnumerator_MoveNext():
                        assay_identifier: str = enumerator_1.System_Collections_Generic_IEnumerator_1_get_Current()
                        with get_enumerator(value_17(assays)) as enumerator_2:
                            while enumerator_2.System_Collections_IEnumerator_MoveNext():
                                assay: ArcAssay = enumerator_2.System_Collections_Generic_IEnumerator_1_get_Current()
                                if assay.Identifier == assay_identifier:
                                    (assay.Performers.append(person))

                person.Comments = Person_removeSourceAssayComments(person)
                if is_empty(source_assays):
                    (persons.append(person))


        finally: 
            dispose(enumerator)


    def mapping_2(arg_11: FSharpList[ArcAssay]) -> Array[str]:
        def mapping_1(a_1: ArcAssay, arg_11: Any=arg_11) -> str:
            return a_1.Identifier

        return list(map_2(mapping_1, arg_11))

    assay_identifiers: Array[str] | None = map(mapping_2, assays)
    def _arrow3257(__unit: None=None) -> str | None:
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("title", string)

    def _arrow3258(__unit: None=None) -> str | None:
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("description", string)

    def _arrow3259(__unit: None=None) -> str | None:
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("submissionDate", string)

    def _arrow3260(__unit: None=None) -> str | None:
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("publicReleaseDate", string)

    def _arrow3261(__unit: None=None) -> Array[Publication] | None:
        arg_21: Decoder_1[Array[Publication]] = resize_array(ISAJson_decoder_4)
        object_arg_9: IOptionalGetter = get.Optional
        return object_arg_9.Field("publications", arg_21)

    def _arrow3262(__unit: None=None) -> Array[OntologyAnnotation] | None:
        arg_23: Decoder_1[Array[OntologyAnnotation]] = resize_array(OntologyAnnotation_ISAJson_decoder)
        object_arg_10: IOptionalGetter = get.Optional
        return object_arg_10.Field("studyDesignDescriptors", arg_23)

    def _arrow3263(__unit: None=None) -> Array[Comment] | None:
        arg_25: Decoder_1[Array[Comment]] = resize_array(ISAJson_decoder_5)
        object_arg_11: IOptionalGetter = get.Optional
        return object_arg_11.Field("comments", arg_25)

    return (ArcStudy(identifier, _arrow3257(), _arrow3258(), _arrow3259(), _arrow3260(), _arrow3261(), None if (len(persons) == 0) else persons, _arrow3262(), tables, None, assay_identifiers, _arrow3263()), default_arg(assays, empty()))


ISAJson_decoder: Decoder_1[tuple[ArcStudy, FSharpList[ArcAssay]]] = Decode_objectNoAdditionalProperties(ISAJson_allowedFields, _arrow3264)

__all__ = ["Helper_getAssayInformation", "encoder", "decoder", "encoder_compressed", "decoder_compressed", "ROCrate_genID", "ROCrate_encoder", "ROCrate_decoder", "ISAJson_encoder", "ISAJson_allowedFields", "ISAJson_decoder"]




from __future__ import annotations
from typing import Any
from ..fable_library.option import default_arg
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.types import Array
from ..fable_library.util import safe_hash
from .cwltypes import CWLType
from .hash_helpers import (box_hash_array, box_hash_seq, hash_1, box_hash_option)

def _expr516() -> TypeInfo:
    return class_type("ARCtrl.CWL.CWLParameterReference", None, CWLParameterReference)


class CWLParameterReference:
    def __init__(self, key: str, values: Array[str] | None=None, type_: CWLType | None=None) -> None:
        self._key: str = key
        self._values: Array[str] = default_arg(values, [])
        self._type: CWLType | None = type_

    @property
    def Key(self, __unit: None=None) -> str:
        this: CWLParameterReference = self
        return this._key

    @Key.setter
    def Key(self, value: str) -> None:
        this: CWLParameterReference = self
        this._key = value

    @property
    def Values(self, __unit: None=None) -> Array[str]:
        this: CWLParameterReference = self
        return this._values

    @Values.setter
    def Values(self, value: Array[str]) -> None:
        this: CWLParameterReference = self
        this._values = value

    @property
    def Type(self, __unit: None=None) -> CWLType | None:
        this: CWLParameterReference = self
        return this._type

    @Type.setter
    def Type(self, value: CWLType | None=None) -> None:
        this: CWLParameterReference = self
        this._type = value

    def __hash__(self, __unit: None=None) -> Any:
        this: CWLParameterReference = self
        return box_hash_array([box_hash_seq(this.Values), hash_1(this.Key), box_hash_option(this.Type)])

    def __eq__(self, obj: Any=None) -> bool:
        this: CWLParameterReference = self
        return this.StructurallyEquals(obj)

    def StructurallyEquals(self, other: CWLParameterReference) -> bool:
        this: CWLParameterReference = self
        return safe_hash(this) == safe_hash(other)

    def ReferenceEquals(self, other: CWLParameterReference) -> bool:
        this: CWLParameterReference = self
        return this is other


CWLParameterReference_reflection = _expr516

def CWLParameterReference__ctor_Z6E62A082(key: str, values: Array[str] | None=None, type_: CWLType | None=None) -> CWLParameterReference:
    return CWLParameterReference(key, values, type_)


__all__ = ["CWLParameterReference_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ....arctrl_core.value import Value
from ....arctrl_json.encode import default_spaces
from ....arctrl_json.Process.value import (decoder as decoder_1, encoder)
from ....fable_library.result import FSharpResult_2
from ....fable_library.string_ import (to_text, printf)
from ....thoth_json_core.types import IEncodable
from ....thoth_json_python.decode import Decode_fromString
from ....thoth_json_python.encode import to_string

def ARCtrl_Value__Value_fromISAJsonString_Static_Z721C83C5(s: str) -> Value:
    match_value: FSharpResult_2[Value, str] = Decode_fromString(decoder_1, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_Value__Value_toISAJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[Value], str]:
    def _arrow3803(v: Value, spaces: Any=spaces) -> str:
        value: IEncodable = encoder(None, v)
        return to_string(default_spaces(spaces), value)

    return _arrow3803


__all__ = ["ARCtrl_Value__Value_fromISAJsonString_Static_Z721C83C5", "ARCtrl_Value__Value_toISAJsonString_Static_71136F3F"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.arc_types import ArcWorkflow
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Table.composite_cell import CompositeCell
from ...arctrl_json.encode import default_spaces
from ...arctrl_json.workflow import (decoder as decoder_1, encoder, decoder_compressed, encoder_compressed)
from ...fable_library.option import default_arg
from ...fable_library.result import FSharpResult_2
from ...fable_library.string_ import (to_text, printf, to_fail)
from ...fable_library.types import Array
from ...thoth_json_core.types import (IEncodable, Decoder_1)
from ...thoth_json_python.decode import Decode_fromString
from ...thoth_json_python.encode import to_string
from .Table.compression import (decode, encode)

def ARCtrl_ArcWorkflow__ArcWorkflow_fromJsonString_Static_Z721C83C5(s: str) -> ArcWorkflow:
    match_value: FSharpResult_2[ArcWorkflow, str] = Decode_fromString(decoder_1, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_ArcWorkflow__ArcWorkflow_toJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[ArcWorkflow], str]:
    def _arrow3859(obj: ArcWorkflow, spaces: Any=spaces) -> str:
        value: IEncodable = encoder(obj)
        return to_string(default_spaces(spaces), value)

    return _arrow3859


def ARCtrl_ArcWorkflow__ArcWorkflow_ToJsonString_71136F3F(this: ArcWorkflow, spaces: int | None=None) -> str:
    return ARCtrl_ArcWorkflow__ArcWorkflow_toJsonString_Static_71136F3F(spaces)(this)


def ARCtrl_ArcWorkflow__ArcWorkflow_fromCompressedJsonString_Static_Z721C83C5(s: str) -> ArcWorkflow:
    try: 
        match_value: FSharpResult_2[ArcWorkflow, str] = Decode_fromString(decode(decoder_compressed), s)
        if match_value.tag == 1:
            raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

        else: 
            return match_value.fields[0]


    except Exception as e_1:
        arg_1: str = str(e_1)
        return to_fail(printf("Error. Unable to parse json string to ArcAssay: %s"))(arg_1)



def ARCtrl_ArcWorkflow__ArcWorkflow_toCompressedJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[ArcWorkflow], str]:
    def _arrow3860(obj: ArcWorkflow, spaces: Any=spaces) -> str:
        return to_string(default_arg(spaces, 0), encode(encoder_compressed, obj))

    return _arrow3860


def ARCtrl_ArcWorkflow__ArcWorkflow_ToCompressedJsonString_71136F3F(this: ArcWorkflow, spaces: int | None=None) -> str:
    return ARCtrl_ArcWorkflow__ArcWorkflow_toCompressedJsonString_Static_71136F3F(spaces)(this)


__all__ = ["ARCtrl_ArcWorkflow__ArcWorkflow_fromJsonString_Static_Z721C83C5", "ARCtrl_ArcWorkflow__ArcWorkflow_toJsonString_Static_71136F3F", "ARCtrl_ArcWorkflow__ArcWorkflow_ToJsonString_71136F3F", "ARCtrl_ArcWorkflow__ArcWorkflow_fromCompressedJsonString_Static_Z721C83C5", "ARCtrl_ArcWorkflow__ArcWorkflow_toCompressedJsonString_Static_71136F3F", "ARCtrl_ArcWorkflow__ArcWorkflow_ToCompressedJsonString_71136F3F"]

